from __future__ import annotations

import json
import logging
import os
import pathlib
import platform
import re
import subprocess
import sys
import typing
from io import TextIOWrapper

from packaging.requirements import Requirement
from packaging.version import Version

from . import dependencies, external_commands, metrics, resolver
from .requirements_file import RequirementType

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)

# uv has no API, so parse its output looking for what it couldn't install.
# Verbose regular expressions ignore blank spaces, so we have to escape those to
# have them recognized as being part of the pattern.
_uv_missing_dependency_pattern = re.compile(
    r"""(
    Because\ there\ is\ no\ version\ of\ (\w+)
    |
    Because\ you\ require\ (\w+)
    |
    Because\ (\w+)\ was\ not\ found\ in\ the\ package\ registry
    )""",
    flags=re.VERBOSE,
)


class MissingDependency(Exception):  # noqa: N818
    def __init__(
        self,
        ctx: context.WorkContext,
        req_type: RequirementType,
        req: Requirement | str | None,
        all_reqs: typing.Iterable[Requirement],
    ):
        self.all_reqs = all_reqs
        resolutions = []
        for r in all_reqs:
            try:
                _, version = resolver.resolve(
                    ctx=ctx,
                    req=r,
                    sdist_server_url=resolver.PYPI_SERVER_URL,
                    req_type=req_type,
                )
            except Exception as err:
                resolutions.append(f"{r} -> {err}")
            else:
                resolutions.append(f"{r} -> {version}")
        formatted_reqs = "\n".join(resolutions)
        if req:
            msg = (
                f"Failed to install {req_type} dependency {req}. "
                f"Check all {req_type} dependencies:\n{formatted_reqs}"
            )
        else:
            msg = (
                f"Failed to install {req_type} dependency. "
                f"Check all {req_type} dependencies:\n{formatted_reqs}"
            )
        super().__init__(f"\n{'*' * 40}\n{msg}\n{'*' * 40}\n")


class BuildEnvironment:
    """Wrapper for a virtualenv used for build isolation"""

    def __init__(
        self,
        ctx: context.WorkContext,
        parent_dir: pathlib.Path,
    ):
        self._ctx = ctx
        self.path = parent_dir.absolute() / f"build-{platform.python_version()}"
        self._createenv()

    @property
    def python(self) -> pathlib.Path:
        """Path to Python interpreter in virtual env"""
        return self.path / "bin" / "python3"

    def get_venv_environ(
        self, template_env: dict[str, str] | None = None
    ) -> dict[str, str]:
        """Add virtual env to extra environ

        The method activates the virtualenv

        1. Put the build environment at the front of the :envvar:`PATH`
           to ensure any build tools are picked up from there and not global
           versions. If the caller has already set a path, start there.
        2. Set :envvar:`VIRTUAL_ENV` so tools looking for that (for example,
           maturin) find it.
        3. Set ``uv`` env vars.
        """
        venv_environ: dict[str, str] = {
            "VIRTUAL_ENV": str(self.path),
        }

        # pre-pend virtualenv's bin
        # 1) template_env PATH, 2) process' PATH, 3) default: "/usr/bin"
        envpath: str | None = None
        if template_env:
            envpath = template_env.get("PATH")
        if envpath is None:
            envpath = os.environ.get("PATH", "/usr/bin")
        envpath_list = envpath.split(os.pathsep)
        venv_bin = str(self.path / "bin")
        if envpath_list[0] != venv_bin:
            envpath_list.insert(0, venv_bin)
        venv_environ["PATH"] = os.pathsep.join(envpath_list)

        # uv env vars, https://docs.astral.sh/uv/reference/environment
        # custom cache for hard links and isolation from user's cache.
        venv_environ["UV_CACHE_DIR"] = str(self._ctx.uv_cache)
        # use system's trust store
        venv_environ["UV_NATIVE_TLS"] = "true"
        # don't use managed Python interpreters
        venv_environ["UV_NO_MANAGED_PYTHON"] = "true"
        venv_environ["UV_PYTHON_DOWNLOADS"] = "never"
        venv_environ["UV_PYTHON"] = str(self.python)

        return venv_environ

    def run(
        self,
        cmd: typing.Sequence[str],
        *,
        cwd: str | None = None,
        extra_environ: dict[str, str] | None = None,
        network_isolation: bool | None = None,
        log_filename: str | None = None,
        stdin: TextIOWrapper | None = None,
    ) -> str:
        """Run command in a virtual environment

        `network_isolation` defaults to context setting.
        """
        extra_environ = extra_environ.copy() if extra_environ else {}
        extra_environ.update(self.get_venv_environ(template_env=extra_environ))

        # default from context
        if network_isolation is None:
            network_isolation = self._ctx.network_isolation
        if network_isolation:
            # Build Rust dependencies without network access
            extra_environ.setdefault("CARGO_NET_OFFLINE", "true")

        return external_commands.run(
            cmd,
            cwd=cwd,
            extra_environ=extra_environ,
            network_isolation=network_isolation,
            log_filename=log_filename,
            stdin=stdin,
        )

    def _createenv(self) -> None:
        if self.path.exists():
            logger.info("reusing build environment in %s", self.path)
            return

        logger.debug("creating build environment in %s", self.path)

        cmd = [
            "uv",
            "venv",
            "--verbose",
            "--python",
            sys.executable,
            "--no-project",
            str(self.path),
        ]
        external_commands.run(
            cmd,
            network_isolation=self._ctx.network_isolation,
        )
        logger.info("created build environment in %s", self.path)

    def install(self, reqs: typing.Iterable[Requirement]) -> None:
        if not reqs:
            return

        # UV does not compile byte code by default
        cmd = [
            "uv",
            "pip",
            "install",
            "--verbose",
            "--upgrade",
            "--only-binary",
            ":all:",
        ]
        cmd.extend(self._ctx.pip_constraint_args)
        cmd.extend(self._ctx.pip_wheel_server_args)
        cmd.extend(str(req) for req in reqs)

        self.run(
            cmd,
            cwd=str(self.path.parent),
            network_isolation=False,
        )
        logger.info(
            "installed dependencies %s into build environment in %s",
            reqs,
            self.path,
        )

    def get_sysconfig_path(self, name: str) -> str:
        """Get ``sysconfig.get_path(name)`` in environment"""
        if not name.isidentifier():
            raise ValueError(f"{name!r} is not an identifier")
        return self.run(
            [
                str(self.python),
                "-c",
                f"import sysconfig; print(sysconfig.get_path('{name}'), end='')",
            ]
        )

    def get_distributions(self) -> typing.Mapping[str, Version]:
        """Get a mapping of installed package names and their versions in the build env"""
        lines = [
            "import sys,json",
            "from importlib.metadata import distributions",
            "json.dump({d.name: d.version for d in distributions()}, sys.stdout)",
        ]
        result = self.run([str(self.python), "-c", "; ".join(lines)])
        try:
            mapping = json.loads(result.strip())
        except Exception:
            logger.exception("failed to de-serialize JSON: %s", result)
        return {name: Version(version) for name, version in sorted(mapping.items())}


@metrics.timeit(description="prepare build environment")
def prepare_build_environment(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
) -> BuildEnvironment:
    logger.info("preparing build environment")

    build_env = BuildEnvironment(
        ctx=ctx,
        parent_dir=sdist_root_dir.parent,
    )

    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_system_dependencies,
        dep_req_type=RequirementType.BUILD_SYSTEM,
    )

    build_backend_dependencies = dependencies.get_build_backend_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_backend_dependencies,
        dep_req_type=RequirementType.BUILD_BACKEND,
    )

    build_sdist_dependencies = dependencies.get_build_sdist_dependencies(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    _safe_install(
        ctx=ctx,
        req=req,
        build_env=build_env,
        deps=build_sdist_dependencies,
        dep_req_type=RequirementType.BUILD_SDIST,
    )

    try:
        distributions = build_env.get_distributions()
    except Exception:
        # ignore error for debug call, error reason is logged in get_distributions()
        pass
    else:
        logger.debug("build env %r has packages %r", build_env.path, distributions)

    return build_env


def _safe_install(
    ctx: context.WorkContext,
    req: Requirement,
    build_env: BuildEnvironment,
    deps: typing.Iterable[Requirement],
    dep_req_type: RequirementType,
) -> None:
    if not deps:
        return
    logger.debug("installing %s %s", dep_req_type, deps)
    try:
        build_env.install(deps)
    except subprocess.CalledProcessError as err:
        logger.error(
            f"{req.name}: failed to install {dep_req_type} dependencies {deps}: {err}"
        )
        match = _uv_missing_dependency_pattern.search(err.output)
        if match is not None:
            req_info = match.groups()[0]
        else:
            req_info = None
        raise MissingDependency(
            ctx,
            dep_req_type,
            req_info,
            deps,
        ) from err

    logger.info("%s: installed %s requirements %s", req.name, dep_req_type, deps)



import importlib.metadata

import click


def _load_commands() -> list[click.Command]:
    """Load commands from fromager.cli entry points"""
    commands: list[click.Command] = []
    seen: dict[str, str] = {}

    for ep in importlib.metadata.entry_points(group="fromager.cli"):
        try:
            command: click.Command | object = ep.load()
        except Exception as e:
            raise RuntimeError(
                f"Unable to load 'fromager.cli' entry point {ep.value!r}"
            ) from e

        # target must be a click command
        if not isinstance(command, click.Command):
            raise RuntimeError(f"{ep.value!r} is not a click.Command: {command}")

        # command name and entry point name have to match
        if command.name != ep.name:
            raise ValueError(
                f"Command name {command.name!r} does not match entry "
                f"point name {ep.name!r} for {ep.value!r}"
            )

        # third party commands can conflict
        if command.name in seen:
            raise ValueError(
                f"Conflict: {ep.value!r} and {seen[ep.name]!r} define {ep.name!r}"
            )
        commands.append(command)
        seen[ep.name] = ep.value

    return commands


commands = _load_commands()



import functools
import itertools
import pathlib
import shutil
import textwrap
import typing
from unittest.mock import Mock, patch

import pytest
from packaging.metadata import Metadata
from packaging.requirements import Requirement
from packaging.utils import NormalizedName
from packaging.version import Version

from fromager import build_environment, context, dependencies

_fromager_root = pathlib.Path(__file__).parent.parent


@pytest.mark.parametrize(
    "build_system,expected_results",
    [
        # Empty
        ({}, dependencies._DEFAULT_BACKEND),
        # Only specify requirements (pyarrow)
        (
            {"requires": ["a-dep"]},
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
        ),
        # Specify everything
        (
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
            {
                "build-backend": "setuptools.build_meta:__legacy__",
                "backend-path": None,
                "requires": ["a-dep"],
            },
        ),
    ],
)
def test_get_build_backend(
    build_system: dict[str, list[str]] | dict[str, str | list[str] | None],
    expected_results: dict[str, typing.Any] | dict[str, str | list[str] | None],
):
    pyproject_toml = {"build-system": build_system}
    actual = dependencies.get_build_backend(pyproject_toml)
    assert expected_results == actual


def _clean_build_artifacts(f):
    @functools.wraps(f)
    def _with_cleanup(*args, **kwds):
        try:
            f(*args, **kwds)
        finally:
            for d in itertools.chain(
                _fromager_root.glob("fromager-*.dist-info"),
                _fromager_root.glob("fromager.egg-info"),
            ):
                shutil.rmtree(d)

    return _with_cleanup


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_system_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    pyproject_file = _fromager_root / "pyproject.toml"
    shutil.copyfile(pyproject_file, tmp_path / "pyproject.toml")

    results = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=tmp_path,
    )
    names = set(r.name for r in results)
    assert names == set(["hatchling", "hatch-vcs"])


def test_get_build_system_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-system-requirements.txt"
    req_file.write_text("foo==1.0")
    results = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=sdist_root_dir,
    )
    assert results == set([Requirement("foo==1.0")])


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_backend_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    # We have to install the build system dependencies into the build
    # environment to get the build sdist dependencies, and we are not running
    # our own local wheel server, so use the public one.
    tmp_context.wheel_server_url = "https://pypi.org/simple"

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
    )
    build_env.install(build_system_dependencies)

    results = dependencies.get_build_backend_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
        build_env=build_env,
    )
    names = set(r.name for r in results)
    assert names == set()


def test_get_build_backend_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-backend-requirements.txt"
    req_file.write_text("foo==1.0")

    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    results = dependencies.get_build_backend_dependencies(
        ctx=tmp_context,
        req=Requirement("fromager"),
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    assert results == set([Requirement("foo==1.0")])


@patch("fromager.dependencies._write_requirements_file")
@_clean_build_artifacts
def test_get_build_sdist_dependencies(
    _: Mock, tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    # We have to install the build system dependencies into the build
    # environment to get the build sdist dependencies, and we are not running
    # our own local wheel server, so use the public one.
    tmp_context.wheel_server_url = "https://pypi.org/simple"

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    build_system_dependencies = dependencies.get_build_system_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
    )
    build_env.install(build_system_dependencies)

    results = dependencies.get_build_sdist_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=_fromager_root,
        build_env=build_env,
    )
    names = set(r.name for r in results)
    assert names == set()


def test_get_build_sdist_dependencies_cached(
    tmp_context: context.WorkContext, tmp_path: pathlib.Path
):
    sdist_root_dir = tmp_path / "sdist"
    sdist_root_dir.mkdir()

    req_file = tmp_path / "build-sdist-requirements.txt"
    req_file.write_text("foo==1.0")

    req = Requirement("fromager")
    build_env = build_environment.BuildEnvironment(
        ctx=tmp_context,
        parent_dir=tmp_path,
    )
    results = dependencies.get_build_sdist_dependencies(
        ctx=tmp_context,
        req=req,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    assert results == set([Requirement("foo==1.0")])


@patch("fromager.dependencies.pep517_metadata_of_sdist")
def test_default_get_install_dependencies_of_sdist(
    m_pep517_metadata_of_sdist: Mock,
    tmp_context: context.WorkContext,
    tmp_path: pathlib.Path,
    caplog: pytest.LogCaptureFixture,
) -> None:
    req = Requirement("huggingface-hub")
    version = Version("1.2.3")
    # sdist metadata name may not be normalized
    metadata_txt = textwrap.dedent(
        """\
        Metadata-Version: 2.3
        Name: HuggingFace_Hub
        Version: 1.2.3
        Requires-Dist: filelock
        Requires-Dist: requests
        """
    )
    metadata = Metadata.from_email(metadata_txt)
    m_pep517_metadata_of_sdist.return_value = metadata

    requirements = dependencies.default_get_install_dependencies_of_sdist(
        ctx=tmp_context,
        req=req,
        version=version,
        sdist_root_dir=tmp_path,
        build_env=Mock(),
        extra_environ={},
        build_dir=tmp_path,
        config_settings={},
    )
    assert requirements == {Requirement("filelock"), Requirement("requests")}

    # bad metadata (version mismatch) is treated as non-fatal error
    metadata_txt = textwrap.dedent(
        """\
        Metadata-Version: 2.3
        Name: HuggingFace_Hub
        Version: 1.2a0
        Requires-Dist: filelock
        Requires-Dist: requests
        """
    )
    metadata = Metadata.from_email(metadata_txt)
    m_pep517_metadata_of_sdist.return_value = metadata
    requirements = dependencies.default_get_install_dependencies_of_sdist(
        ctx=tmp_context,
        req=req,
        version=version,
        sdist_root_dir=tmp_path,
        build_env=Mock(),
        extra_environ={},
        build_dir=tmp_path,
        config_settings={},
    )
    assert requirements == {Requirement("filelock"), Requirement("requests")}
    assert (
        "sdist metadata '1.2a0' does not match public version '1.2.3'"
        in caplog.messages[-1]
    )


@pytest.mark.parametrize(
    "req_str,version_str,dist_name_str,dist_version_str,exc",
    [
        ("mypkg", "1.0", "mypkg", "1.0", None),
        ("MyPKG", "1.0", "mypkg", "1.0", None),
        ("mypkg", "1.0", "MyPKG", "1.0", RuntimeError),
        ("mypkg", "1.0", "otherpkg", "1.0", ValueError),
        ("mypkg", "1.0", "mypkg", "1.1", ValueError),
        ("mypkg", "1.0+local", "mypkg", "1.0+local", None),
        ("mypkg", "1.0", "mypkg", "1.0+local", None),
        ("mypkg", "1.0+local", "mypkg", "1.0", None),
    ],
)
def test_validate_dist_name_version(
    req_str: str,
    version_str: str,
    dist_name_str: str,
    dist_version_str: str,
    exc: type[Exception] | None,
) -> None:
    validate = functools.partial(
        dependencies.validate_dist_name_version,
        req=Requirement(req_str),
        version=Version(version_str),
        what="test",
        dist_name=typing.cast(NormalizedName, dist_name_str),
        dist_version=Version(dist_version_str),
    )
    if exc is None:
        validate()
    else:
        with pytest.raises(exc):
            validate()



import typing

import click

from fromager.commands import bootstrap, build


def get_option_names(cmd: click.Command) -> typing.Iterable[str]:
    return [o.name for o in cmd.params if o.name]


def test_bootstrap_pallel_options() -> None:
    expected: set[str] = set()
    expected.update(get_option_names(bootstrap.bootstrap))
    expected.update(get_option_names(build.build_parallel))
    # bootstrap-parallel enforces sdist_only=True and handles
    # graph_file internally.
    expected.discard("sdist_only")
    expected.discard("graph_file")

    assert set(get_option_names(bootstrap.bootstrap_parallel)) == expected



import concurrent.futures
import dataclasses
import datetime
import json
import logging
import pathlib
import sys
import threading
import typing
from urllib.parse import urlparse

import click
import rich
import rich.box
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name, parse_wheel_filename
from packaging.version import Version
from rich.table import Table
from rich.text import Text

from fromager import (
    build_environment,
    clickext,
    context,
    dependency_graph,
    hooks,
    metrics,
    overrides,
    progress,
    read,
    server,
    sources,
    wheels,
)

from .. import resolver
from ..log import VERBOSE_LOG_FMT, ThreadLogFilter, req_ctxvar_context

logger = logging.getLogger(__name__)

DependencyNodeList = list[dependency_graph.DependencyNode]


@dataclasses.dataclass(order=True, frozen=True)
class BuildSequenceEntry:
    # compare, hash, and sort by name and version
    name: str
    version: Version
    prebuilt: bool = dataclasses.field(compare=False)
    download_url: str = dataclasses.field(compare=False)
    wheel_filename: pathlib.Path = dataclasses.field(compare=False)
    skipped: bool = dataclasses.field(default=False, compare=False)

    @staticmethod
    def dict_factory(x):
        return {
            k: str(v) if isinstance(v, pathlib.Path | Version) else v for (k, v) in x
        }


@click.command()
@click.option(
    "--wheel-server-url",
    default="",
    type=str,
    help="URL for the wheel server for builds",
)
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.argument("sdist_server_url")
@click.pass_obj
def build(
    wkctx: context.WorkContext,
    wheel_server_url: str,
    dist_name: str,
    dist_version: Version,
    sdist_server_url: str,
) -> None:
    """Build a single version of a single wheel

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    1. Downloads the source distribution.

    2. Unpacks it and prepares the source via patching, vendoring rust
       dependencies, etc.

    3. Prepares a build environment with the build dependencies.

    4. Builds the wheel.

    Refer to the 'step' commands for scripting these stages
    separately.

    """
    wkctx.wheel_server_url = wheel_server_url
    server.start_wheel_server(wkctx)
    req = Requirement(f"{dist_name}=={dist_version}")
    with req_ctxvar_context(req, dist_version):
        # We have to resolve the source here to get a
        # source_url. Other build modes use data computed from a
        # bootstrap job where that URL is saved in the build
        # instruction file passed to build-sequence or build-parallel.
        source_url, version = sources.resolve_source(
            ctx=wkctx,
            req=req,
            sdist_server_url=sdist_server_url,
        )
        entry = _build(
            wkctx=wkctx,
            resolved_version=version,
            req=req,
            source_download_url=source_url,
            force=True,
            cache_wheel_server_url=None,
        )
    print(entry.wheel_filename)


build._fromager_show_build_settings = True  # type: ignore


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="rebuild wheels even if they have already been built",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can check if it had already built the wheel",
)
@click.argument("build_order_file")
@click.pass_obj
def build_sequence(
    wkctx: context.WorkContext,
    build_order_file: str,
    force: bool,
    cache_wheel_server_url: str | None,
) -> None:
    """Build a sequence of wheels in order

    BUILD_ORDER_FILE is the build-order.json files to build

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    Performs the equivalent of the 'build' command for each item in
    the build order file.

    """
    server.start_wheel_server(wkctx)

    if force:
        logger.info(
            "rebuilding all wheels even if they exist in "
            f"{wkctx.wheel_server_url=}, {cache_wheel_server_url=}"
        )
    else:
        logger.info(
            "skipping builds for versions of packages available at "
            f"{wkctx.wheel_server_url=}, {cache_wheel_server_url=}"
        )

    entries: list[BuildSequenceEntry] = []

    logger.info("reading build order from %s", build_order_file)
    with read.open_file_or_url(build_order_file) as f:
        for entry in progress.progress(json.load(f)):
            dist_name = entry["dist"]
            resolved_version = Version(entry["version"])
            source_download_url = entry["source_url"]

            # If we are building from git, use the requirement as specified so
            # we include the URL. Otherwise, create a fake requirement with the
            # name and version so we are explicitly building the expected
            # version.
            if entry["source_url_type"] == "git":
                req = Requirement(entry["req"])
            else:
                req = Requirement(f"{dist_name}=={resolved_version}")

            with req_ctxvar_context(req, resolved_version):
                logger.info("building %s", resolved_version)
                entry = _build(
                    wkctx=wkctx,
                    resolved_version=resolved_version,
                    req=req,
                    source_download_url=source_download_url,
                    force=force,
                    cache_wheel_server_url=cache_wheel_server_url,
                )
                if entry.prebuilt:
                    logger.info(
                        "downloaded prebuilt wheel %s", entry.wheel_filename.name
                    )
                elif entry.skipped:
                    logger.info(
                        "skipping building wheel since %s already exists",
                        entry.wheel_filename.name,
                    )
                else:
                    logger.info("built %s", entry.wheel_filename.name)

                entries.append(entry)

    metrics.summarize(wkctx, "Building")

    _summary(wkctx, entries)


build_sequence._fromager_show_build_settings = True  # type: ignore


def _summary(ctx: context.WorkContext, entries: list[BuildSequenceEntry]) -> None:
    output: list[typing.Any] = []
    now = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%SZ")
    output.append(Text(f"Build sequence summary {now}\n"))

    built_entries = [e for e in entries if not e.skipped and not e.prebuilt]
    if built_entries:
        output.append(
            _create_table(
                built_entries,
                title="New builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No new builds\n"))

    prebuilt_entries = [e for e in entries if e.prebuilt]
    if prebuilt_entries:
        output.append(
            _create_table(
                prebuilt_entries,
                title="Prebuilt wheels",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No pre-built wheels\n"))

    skipped_entries = [e for e in entries if e.skipped and not e.prebuilt]
    if skipped_entries:
        output.append(
            _create_table(
                skipped_entries,
                title="Skipped existing builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No skipped builds\n"))

    console = rich.get_console()
    console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.md", "w", encoding="utf-8") as f:
        console = rich.console.Console(file=f, width=sys.maxsize)
        console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.json", "w", encoding="utf-8") as f:
        json.dump(
            [
                dataclasses.asdict(e, dict_factory=BuildSequenceEntry.dict_factory)
                for e in entries
            ],
            f,
        )


def _create_table(entries: list[BuildSequenceEntry], **table_kwargs) -> Table:
    table = Table(**table_kwargs)
    table.add_column("Name", justify="right", no_wrap=True)
    table.add_column("Version", no_wrap=True)
    table.add_column("Wheel", no_wrap=True)
    table.add_column("Source URL")

    platlib_count = 0

    for info in sorted(entries):
        tags = parse_wheel_filename(info.wheel_filename.name)[3]
        if any(t.platform != "any" or t.abi != "none" for t in tags):
            platlib_count += 1
        source_filename = urlparse(info.download_url).path.rsplit("/", 1)[-1]
        table.add_row(
            info.name,
            str(info.version),
            info.wheel_filename.name,
            # escape Rich markup
            rf"\[{source_filename}]({info.download_url})",
        )

    # summary
    table.add_section()
    table.add_row(
        f"total: {len(entries)}",
        None,
        f"platlib: {platlib_count}",
        None,
    )
    return table


def _build(
    wkctx: context.WorkContext,
    resolved_version: Version,
    req: Requirement,
    source_download_url: str,
    force: bool,
    cache_wheel_server_url: str | None,
) -> BuildSequenceEntry:
    """Handle one version of one wheel.

    Either:
    1. Reuse an existing wheel we have locally.
    2. Download a pre-built wheel.
    3. Build the wheel from source.
    """
    wheel_filename: pathlib.Path | None = None
    use_exiting_wheel: bool = False

    # Set up a log file for all of the details of the build for this one wheel.
    # We attach a handler to the root logger so that all messages are logged to
    # the file, and we add a filter to the handler so that only messages from
    # the current thread are logged for when we build in parallel.
    root_logger = logging.getLogger(None)
    module_name = overrides.pkgname_to_override_module(req.name)
    wheel_log = wkctx.logs_dir / f"{module_name}-{resolved_version}.log"
    file_handler = logging.FileHandler(filename=str(wheel_log))
    file_handler.setFormatter(logging.Formatter(VERBOSE_LOG_FMT))
    file_handler.addFilter(ThreadLogFilter(threading.current_thread().name))
    root_logger.addHandler(file_handler)

    logger.info("starting processing")
    pbi = wkctx.package_build_info(req)
    prebuilt = pbi.pre_built

    wheel_server_urls = wheels.get_wheel_server_urls(
        wkctx, req, cache_wheel_server_url=cache_wheel_server_url
    )

    # See if we can reuse an existing wheel.
    if not force:
        wheel_filename = _is_wheel_built(
            wkctx,
            req.name,
            resolved_version,
            wheel_server_urls,
        )
        if wheel_filename:
            logger.info("using existing wheel from %s", wheel_filename)
            use_exiting_wheel = True

    # Handle prebuilt wheels.
    if prebuilt:
        if not wheel_filename:
            logger.info("downloading prebuilt wheel")
            wheel_filename = wheels.download_wheel(
                req=req,
                wheel_url=source_download_url,
                output_directory=wkctx.wheels_build,
            )
        else:
            # already downloaded prebuilt wheel
            use_exiting_wheel = True

        # Run hooks for prebuilt wheels. At this point wheel_filename should
        # be set either from _is_wheel_built() or download_wheel().
        hooks.run_prebuilt_wheel_hooks(
            ctx=wkctx,
            req=req,
            dist_name=req.name,
            dist_version=str(resolved_version),
            wheel_filename=wheel_filename,
        )

    # If we get here and still don't have a wheel filename, then we need to
    # build the wheel.
    if not wheel_filename:
        source_filename = sources.download_source(
            ctx=wkctx,
            req=req,
            version=resolved_version,
            download_url=source_download_url,
        )
        logger.debug(
            "saved sdist of version %s from %s to %s",
            resolved_version,
            source_download_url,
            source_filename,
        )

        # Prepare source
        source_root_dir = sources.prepare_source(
            ctx=wkctx,
            req=req,
            source_filename=source_filename,
            version=resolved_version,
        )

        # Build environment
        build_env = build_environment.prepare_build_environment(
            ctx=wkctx, req=req, sdist_root_dir=source_root_dir
        )

        # Make a new source distribution, in case we patched the code.
        sdist_filename = sources.build_sdist(
            ctx=wkctx,
            req=req,
            version=resolved_version,
            sdist_root_dir=source_root_dir,
            build_env=build_env,
        )

        # Build
        wheel_filename = wheels.build_wheel(
            ctx=wkctx,
            req=req,
            sdist_root_dir=source_root_dir,
            version=resolved_version,
            build_env=build_env,
        )

        hooks.run_post_build_hooks(
            ctx=wkctx,
            req=req,
            dist_name=canonicalize_name(req.name),
            dist_version=str(resolved_version),
            sdist_filename=sdist_filename,
            wheel_filename=wheel_filename,
        )

        wkctx.clean_build_dirs(source_root_dir, build_env)

    root_logger.removeHandler(file_handler)
    file_handler.close()

    server.update_wheel_mirror(wkctx)

    # After we update the wheel mirror, the built file has
    # moved to a new directory.
    wheel_filename = wkctx.wheels_downloads / wheel_filename.name

    return BuildSequenceEntry(
        name=canonicalize_name(req.name),
        version=resolved_version,
        prebuilt=prebuilt,
        download_url=source_download_url,
        wheel_filename=wheel_filename,
        skipped=use_exiting_wheel,
    )


def _is_wheel_built(
    wkctx: context.WorkContext,
    dist_name: str,
    resolved_version: Version,
    wheel_server_urls: list[str],
) -> pathlib.Path | None:
    req = Requirement(f"{dist_name}=={resolved_version}")

    try:
        logger.info(
            "checking if a suitable wheel for %s was already built on %s",
            req,
            wheel_server_urls,
        )
        url, _ = wheels.resolve_prebuilt_wheel(
            ctx=wkctx,
            req=req,
            wheel_server_urls=wheel_server_urls,
        )
        logger.info("found candidate wheel %s", url)
        pbi = wkctx.package_build_info(req)
        build_tag_from_settings = pbi.build_tag(resolved_version)
        build_tag = build_tag_from_settings if build_tag_from_settings else (0, "")
        wheel_basename = resolver.extract_filename_from_url(url)
        _, _, build_tag_from_name, _ = parse_wheel_filename(wheel_basename)
        existing_build_tag = build_tag_from_name if build_tag_from_name else (0, "")
        if (
            existing_build_tag[0] > build_tag[0]
            and existing_build_tag[1] == build_tag[1]
        ):
            raise ValueError(
                f"{dist_name}: changelog for version {resolved_version} is inconsistent. Found build tag {existing_build_tag} but expected {build_tag}"
            )
        if existing_build_tag != build_tag:
            logger.info(
                f"candidate wheel build tag {existing_build_tag} does not match expected build tag {build_tag}"
            )
            return None

        wheel_filename: pathlib.Path | None = None
        if url.startswith(wkctx.wheel_server_url):
            logging.debug("found wheel on local server")
            wheel_filename = wkctx.wheels_downloads / wheel_basename
            if not wheel_filename.exists():
                logger.info("wheel not found in local cache, preparing to download")
                wheel_filename = None

        if not wheel_filename:
            # if the found wheel was on an external server, then download it
            logger.info("downloading wheel from %s", url)
            wheel_filename = wheels.download_wheel(req, url, wkctx.wheels_downloads)

        return wheel_filename
    except Exception:
        logger.debug(
            "could not locate prebuilt wheel %s-%s on %s",
            dist_name,
            resolved_version,
            wheel_server_urls,
            exc_info=True,
        )
        logger.info("could not locate prebuilt wheel")
        return None


def _build_parallel(
    wkctx: context.WorkContext,
    resolved_version: Version,
    req: Requirement,
    source_download_url: str,
    force: bool,
    cache_wheel_server_url: str | None,
) -> BuildSequenceEntry:
    """
    This function runs in a thread to manage the build of a single package.
    """
    with req_ctxvar_context(req, resolved_version):
        return _build(
            wkctx=wkctx,
            resolved_version=resolved_version,
            req=req,
            source_download_url=source_download_url,
            force=force,
            cache_wheel_server_url=cache_wheel_server_url,
        )


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="rebuild wheels even if they have already been built",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can check if it had already built the wheel",
)
@click.option(
    "-m",
    "--max-workers",
    type=int,
    default=None,
    help="maximum number of parallel workers to run (default: unlimited)",
)
@click.argument("graph_file")
@click.pass_obj
def build_parallel(
    wkctx: context.WorkContext,
    graph_file: str,
    force: bool,
    cache_wheel_server_url: str | None,
    max_workers: int | None,
) -> None:
    """Build wheels in parallel based on a dependency graph

    GRAPH_FILE is a graph.json file containing the dependency relationships between packages

    Performs parallel builds of wheels based on their dependency relationships.
    Packages that have no dependencies or whose dependencies are already built
    can be built concurrently. By default, all possible packages are built in
    parallel. Use --max-workers to limit the number of concurrent builds.

    """
    wkctx.enable_parallel_builds()

    server.start_wheel_server(wkctx)
    wheel_server_urls: list[str] = [wkctx.wheel_server_url]
    if cache_wheel_server_url:
        # put after local server so we always check local server first
        wheel_server_urls.append(cache_wheel_server_url)

    if force:
        logger.info(f"rebuilding all wheels even if they exist in {wheel_server_urls}")
    else:
        logger.info(
            f"skipping builds for versions of packages available at {wheel_server_urls}"
        )

    # Load the dependency graph
    logger.info("reading dependency graph from %s", graph_file)
    graph: dependency_graph.DependencyGraph
    graph = dependency_graph.DependencyGraph.from_file(graph_file)

    # Track what has been built
    built_node_keys: set[str] = set()

    # Get all nodes that need to be built (excluding prebuilt ones and the root node)
    # Sort the nodes to build by their key one time to avoid
    # redoing the sort every iteration and to make the output deterministic.
    nodes_to_build: DependencyNodeList = sorted(
        (n for n in graph.nodes.values() if n.key != dependency_graph.ROOT),
        key=lambda n: n.key,
    )
    logger.info("found %d packages to build", len(nodes_to_build))

    # A node can be built when all of its build dependencies are built
    entries: list[BuildSequenceEntry] = []

    with progress.progress_context(total=len(nodes_to_build)) as progressbar:

        def update_progressbar_cb(future: concurrent.futures.Future) -> None:
            """Immediately update the progress when when a task is done"""
            progressbar.update()

        while nodes_to_build:
            # Find nodes that can be built (all build dependencies are built)
            buildable_nodes: DependencyNodeList = []
            for node in nodes_to_build:
                with req_ctxvar_context(
                    Requirement(node.canonicalized_name), node.version
                ):
                    # Get all build dependencies (build-system, build-backend, build-sdist)
                    build_deps: DependencyNodeList = [
                        edge.destination_node
                        for edge in node.children
                        if edge.req_type.is_build_requirement
                    ]
                    # A node can be built when all of its build dependencies are built
                    unbuilt_deps: set[str] = set(
                        dep.key for dep in build_deps if dep.key not in built_node_keys
                    )
                    if not unbuilt_deps:
                        logger.info(
                            "ready to build, have all build dependencies: %s",
                            sorted(set(dep.key for dep in build_deps)),
                        )
                        buildable_nodes.append(node)
                    else:
                        logger.info(
                            "waiting for build dependencies: %s",
                            sorted(unbuilt_deps),
                        )

            if not buildable_nodes:
                # If we can't build anything but still have nodes, we have a cycle
                remaining: list[str] = [n.key for n in nodes_to_build]
                logger.info("have already built: %s", sorted(built_node_keys))
                raise ValueError(f"Circular dependency detected among: {remaining}")

            logger.info(
                "ready to build: %s",
                sorted(n.key for n in buildable_nodes),
            )

            # Check if any buildable node requires exclusive build (exclusive_build == True)
            exclusive_nodes: DependencyNodeList = [
                node
                for node in buildable_nodes
                if wkctx.settings.package_build_info(
                    node.canonicalized_name
                ).exclusive_build
            ]
            if exclusive_nodes:
                # Only build the first exclusive node this round
                buildable_nodes = [exclusive_nodes[0]]
                logger.info(
                    f"{exclusive_nodes[0].canonicalized_name}: requires exclusive build, running it alone this round."
                )

            # Build up to max_workers nodes concurrently (or all if max_workers is None)
            with concurrent.futures.ThreadPoolExecutor(
                max_workers=max_workers
            ) as executor:
                futures: list[concurrent.futures.Future[tuple[pathlib.Path, bool]]] = []
                reqs: list[Requirement] = []
                logger.info(
                    "starting to build: %s", sorted(n.key for n in buildable_nodes)
                )
                for node in buildable_nodes:
                    req = Requirement(f"{node.canonicalized_name}=={node.version}")
                    reqs.append(req)
                    future = executor.submit(
                        _build_parallel,
                        wkctx=wkctx,
                        resolved_version=node.version,
                        req=req,
                        source_download_url=node.download_url,
                        force=force,
                        cache_wheel_server_url=cache_wheel_server_url,
                    )
                    future.add_done_callback(update_progressbar_cb)
                    futures.append(future)

                # Wait for all builds to complete
                for node, future in zip(buildable_nodes, futures, strict=True):
                    try:
                        entry = future.result()
                        entries.append(entry)
                        built_node_keys.add(node.key)
                        nodes_to_build.remove(node)
                        # progress bar is updated in callback
                    except Exception as e:
                        logger.error(f"Failed to build {node.key}: {e}")
                        raise

    metrics.summarize(wkctx, "Building in parallel")
    _summary(wkctx, entries)


build_parallel._fromager_show_build_settings = True  # type: ignore



from __future__ import annotations

import collections
import logging
import os
import pathlib
import shutil
import sys
import tempfile
import textwrap
import typing
import zipfile

import elfdeps
import tomlkit
import wheel.wheelfile  # type: ignore
from packaging.requirements import Requirement
from packaging.tags import Tag
from packaging.utils import (
    BuildTag,
    canonicalize_name,
    parse_wheel_filename,
)
from packaging.version import Version

from . import (
    dependencies,
    external_commands,
    metrics,
    overrides,
    packagesettings,
    requirements_file,
    resolver,
    sources,
)

if typing.TYPE_CHECKING:
    from . import build_environment, context

logger = logging.getLogger(__name__)

FROMAGER_BUILD_SETTINGS = "fromager-build-settings"
FROMAGER_ELF_PROVIDES = "fromager-elf-provides.txt"
FROMAGER_ELF_REQUIRES = "fromager-elf-requires.txt"
FROMAGER_BUILD_REQ_PREFIX = "fromager"


def _extra_metadata_elfdeps(
    ctx: context.WorkContext,
    req: Requirement,
    wheel_root_dir: pathlib.Path,
    dist_info_dir: pathlib.Path,
) -> typing.Iterable[elfdeps.ELFInfo]:
    """Analyze a wheel's ELF dependencies and add info files

    Logs and returns library dependencies and library provides. Writes
    requirements to dist-info.
    """
    # mapping of required libraries to list of versions
    requires: set[elfdeps.SOInfo] = set()
    provides: set[elfdeps.SOInfo] = set()
    runpaths: set[str] = set()
    elfinfos: list[elfdeps.ELFInfo] = []

    settings = elfdeps.ELFAnalyzeSettings(filter_soname=True)
    for info in elfdeps.analyze_dirtree(wheel_root_dir, settings=settings):
        if info.filename is not None:
            relname = str(info.filename.relative_to(wheel_root_dir))
        else:
            relname = "n/a"
        logger.debug(
            f"{relname} ({info.soname}) "
            f"requires {sorted(info.requires)}, "
            f"provides {sorted(info.provides)}"
        )
        provides.update(info.provides)
        requires.update(info.requires)
        if info.runpath is not None:
            runpaths.update(info.runpath)
        elfinfos.append(info)

    # Don't list provided names as requirements
    requires = requires.difference(provides)

    if requires:
        reqmap: dict[str, list[str]] = collections.defaultdict(list)
        for r in requires:
            reqmap[r.soname].append(r.version)

        names = sorted(
            name for name in reqmap if not name.startswith(("ld-linux", "rtld"))
        )
        logger.info("Requires libraries: %s", ", ".join(names))
        for name, versions in sorted(reqmap.items()):
            logger.debug(
                "Requires %s(%s)",
                name,
                ", ".join(v for v in versions if v),
            )

        requires_file = dist_info_dir / FROMAGER_ELF_REQUIRES
        with requires_file.open("w", encoding="utf-8") as f:
            for soinfo in sorted(requires):
                f.write(f"{soinfo}\n")

    if provides:
        names = sorted(p.soname for p in provides)
        logger.info("Provides libraries: %s", ", ".join(names))

        provides_file = dist_info_dir / FROMAGER_ELF_PROVIDES
        with provides_file.open("w", encoding="utf-8") as f:
            for soinfo in sorted(provides):
                f.write(f"{soinfo}\n")

    if runpaths:
        logger.info("Libraries have runpath: %s", " ".join(sorted(runpaths)))

    return elfinfos


def extract_info_from_wheel_file(
    req: Requirement, wheel_file: pathlib.Path
) -> tuple[str, Version, BuildTag, frozenset[Tag]]:
    # parse_wheel_filename normalizes the dist name, however the dist-info
    # directory uses the verbatim distribution name from the wheel file.
    # Packages with upper case names like "MarkupSafe" are affected.
    dist_name_normalized, dist_version, build_tag, wheel_tags = parse_wheel_filename(
        wheel_file.name
    )
    dist_name = wheel_file.name.split("-", 1)[0]
    if dist_name_normalized != canonicalize_name(dist_name):
        # sanity check, should never fail
        raise ValueError(f"{dist_name_normalized} does not match {dist_name}")
    return (dist_name, dist_version, build_tag, wheel_tags)


def default_add_extra_metadata_to_wheels(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    extra_environ: dict[str, str],
    sdist_root_dir: pathlib.Path,
    dist_info_dir: pathlib.Path,
) -> dict[str, typing.Any]:
    raise NotImplementedError


@metrics.timeit(description="add extra metadata to wheels")
def add_extra_metadata_to_wheels(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    extra_environ: dict[str, str],
    sdist_root_dir: pathlib.Path,
    wheel_file: pathlib.Path,
) -> pathlib.Path:
    pbi = ctx.package_build_info(req)
    dist_name, dist_version, _, wheel_tags = extract_info_from_wheel_file(
        req, wheel_file
    )
    dist_filename = f"{dist_name}-{dist_version}"

    extra_data_plugin = overrides.find_override_method(
        req.name, "add_extra_metadata_to_wheels"
    )
    data_to_add = {}

    with tempfile.TemporaryDirectory() as dir_name:
        wheel_root_dir = pathlib.Path(dir_name) / dist_filename
        wheel_root_dir.mkdir()
        with zipfile.ZipFile(str(wheel_file)) as zf:
            for infolist in zf.filelist:
                # Check for path traversal attempts
                if (
                    os.path.isabs(infolist.filename)
                    or ".." in pathlib.Path(infolist.filename).parts
                ):
                    raise ValueError(f"Unsafe path in wheel: {infolist.filename}")
                zf.extract(infolist, wheel_root_dir)
                # the higher 16 bits store the permissions and type of file (i.e. stat.filemode)
                # the lower bits of this give us the permission
                permissions = infolist.external_attr >> 16 & 0o777
                wheel_root_dir.joinpath(infolist.filename).chmod(permissions)

        dist_info_dir = wheel_root_dir / f"{dist_filename}.dist-info"
        if not dist_info_dir.is_dir():
            raise ValueError(f"{wheel_file} does not contain {dist_info_dir.name}")

        if extra_data_plugin:
            data_to_add = overrides.invoke(
                extra_data_plugin,
                ctx=ctx,
                req=req,
                version=version,
                extra_environ=extra_environ,
                sdist_root_dir=sdist_root_dir,
                dist_info_dir=dist_info_dir,
            )
            if not isinstance(data_to_add, dict):
                logger.warning(
                    "unexpected return type from plugin add_extra_metadata_to_wheels. Expected dictionary. Will ignore"
                )
                data_to_add = {}

        if pbi.has_config:
            settings = pbi.serialize(mode="json", exclude_defaults=False)
        else:
            settings = {}
        if data_to_add:
            settings["metadata-from-plugin"] = data_to_add
        build_file = dist_info_dir / FROMAGER_BUILD_SETTINGS
        build_file.write_text(tomlkit.dumps(settings))

        req_files = sdist_root_dir.parent.glob("*-requirements.txt")
        for req_file in req_files:
            shutil.copy(
                req_file, dist_info_dir / f"{FROMAGER_BUILD_REQ_PREFIX}-{req_file.name}"
            )

        if any(tag.platform != "all" for tag in wheel_tags):
            # platlib wheel
            if sys.platform == "linux":
                _extra_metadata_elfdeps(
                    ctx=ctx,
                    req=req,
                    wheel_root_dir=wheel_root_dir,
                    dist_info_dir=dist_info_dir,
                )
            else:
                logger.debug(
                    "shared library dependency analysis not implemented for %s",
                    sys.platform,
                )
        else:
            logger.debug("%s is a purelib wheel", req.name)

        build_tag_from_settings = pbi.build_tag(version)
        build_tag = build_tag_from_settings if build_tag_from_settings else (0, "")

        cmd = [
            "wheel",
            "pack",
            str(wheel_root_dir),
            "--dest-dir",
            str(wheel_file.parent),
            "--build-number",
            f"{build_tag[0]}{build_tag[1]}",
        ]

        external_commands.run(
            cmd,
            cwd=dir_name,
            network_isolation=ctx.network_isolation,
        )

    wheel_file.unlink(missing_ok=True)
    wheels = list(wheel_file.parent.glob(f"{dist_filename}-*.whl"))
    if wheels:
        logger.info(
            f"added extra metadata and build tag {build_tag}, wheel renamed from {wheel_file.name} to {wheels[0].name}"
        )
        return wheels[0]
    raise FileNotFoundError("Could not locate new wheels file")


def validate_wheel_filename(
    req: Requirement,
    version: Version,
    wheel_file: pathlib.Path,
) -> None:
    """Check that wheel matches requirement name and version"""
    wheel_name, wheel_version, _, _ = parse_wheel_filename(wheel_file.name)
    dependencies.validate_dist_name_version(
        req=req,
        version=version,
        what=wheel_file.name,
        dist_name=wheel_name,
        dist_version=wheel_version,
    )


@metrics.timeit(description="build wheels")
def build_wheel(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    pbi = ctx.package_build_info(req)
    logger.info(
        f"building {ctx.variant} wheel for {req} in {sdist_root_dir} "
        f"writing to {ctx.wheels_build}"
    )

    # add package and variant env vars, package's parallel job vars, and
    # build_env's virtual env vars.
    extra_environ = packagesettings.get_extra_environ(
        ctx=ctx,
        req=req,
        sdist_root_dir=sdist_root_dir,
        version=version,
        build_env=build_env,
    )

    if (
        pbi.build_ext_parallel
        and "DIST_EXTRA_CONFIG" not in extra_environ
        and "MAX_JOBS" in extra_environ
    ):
        # configure setuptools to use parallel builds
        # https://setuptools.pypa.io/en/latest/deprecated/distutils/configfile.html
        dist_extra_cfg = build_env.path / "dist-extra.cfg"
        dist_extra_cfg.write_text(
            textwrap.dedent(
                f"""
                [build_ext]
                parallel = {extra_environ["MAX_JOBS"]}
                """
            )
        )
        extra_environ["DIST_EXTRA_CONFIG"] = str(dist_extra_cfg)

    overrides.find_and_invoke(
        req.name,
        "build_wheel",
        default_build_wheel,
        ctx=ctx,
        build_env=build_env,
        extra_environ=extra_environ,
        req=req,
        version=version,
        sdist_root_dir=sdist_root_dir,
        build_dir=pbi.build_dir(sdist_root_dir),
    )

    wheels = list(ctx.wheels_build.glob("*.whl"))
    if len(wheels) != 1:
        raise FileNotFoundError(
            f"Expected 1 built wheel in {ctx.wheels_build}, got {len(wheels)}"
        )
    tmp_wheel_file = wheels[0]

    # validate location and file name
    if tmp_wheel_file.parent != ctx.wheels_build:
        raise ValueError(
            f"{tmp_wheel_file!r} is not in wheels build directory {ctx.wheels_build!r}"
        )
    validate_wheel_filename(req=req, version=version, wheel_file=tmp_wheel_file)

    # add extra metadata and validate again
    new_wheel_file = add_extra_metadata_to_wheels(
        ctx=ctx,
        req=req,
        version=version,
        extra_environ=extra_environ,
        sdist_root_dir=sdist_root_dir,
        wheel_file=tmp_wheel_file,
    )
    validate_wheel_filename(req=req, version=version, wheel_file=new_wheel_file)

    # invalidate uv cache, so the new wheel is picked up. The step is only
    # relevant for local development when a package is rebuilt multiple times
    # without bumping the build tag.
    ctx.uv_clean_cache(req)

    return new_wheel_file


def pep517_build_wheel(
    ctx: context.WorkContext,
    build_env: build_environment.BuildEnvironment,
    extra_environ: dict[str, str],
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_dir: pathlib.Path,
) -> pathlib.Path:
    """Use the PEP 517 API to build a wheel distribution"""
    logger.debug(f"building wheel in {build_dir} with {extra_environ}")

    hook_caller = dependencies.get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
        log_filename=str(sdist_root_dir.parent / "build.log"),
    )

    pbi = ctx.package_build_info(req)
    wheel_filename = hook_caller.build_wheel(
        str(ctx.wheels_build),
        config_settings=pbi.config_settings,
    )
    logger.debug("built wheel %s", wheel_filename)

    return ctx.wheels_build / wheel_filename


def default_build_wheel(
    ctx: context.WorkContext,
    build_env: build_environment.BuildEnvironment,
    extra_environ: dict[str, str],
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_dir: pathlib.Path,
) -> pathlib.Path:
    return pep517_build_wheel(
        ctx=ctx,
        build_env=build_env,
        extra_environ=extra_environ,
        req=req,
        sdist_root_dir=sdist_root_dir,
        version=version,
        build_dir=build_dir,
    )


def download_wheel(
    req: Requirement,
    wheel_url: str,
    output_directory: pathlib.Path,
) -> pathlib.Path:
    wheel_filename = output_directory / resolver.extract_filename_from_url(wheel_url)
    if not wheel_filename.exists():
        logger.info(f"downloading pre-built wheel {wheel_url}")
        wheel_filename = _download_wheel_check(req, output_directory, wheel_url)
        logger.info(f"saved wheel to {wheel_filename}")
    else:
        logger.info(f"have existing wheel {wheel_filename}")

    return wheel_filename


def _download_wheel_check(
    req: Requirement, destination_dir: pathlib.Path, wheel_url: str
) -> pathlib.Path:
    wheel_filename = sources.download_url(
        req=req,
        destination_dir=destination_dir,
        url=wheel_url,
    )
    # validates whether the wheel is correct or not. will raise an error in the wheel is invalid
    wheel.wheelfile.WheelFile(wheel_filename)
    return wheel_filename


def get_wheel_server_urls(
    ctx: context.WorkContext, req: Requirement, *, cache_wheel_server_url: str | None
) -> list[str]:
    pbi = ctx.package_build_info(req)
    wheel_server_urls: list[str] = []
    if pbi.wheel_server_url:
        # use only the wheel server from settings if it is defined. Do not fallback to other URLs
        wheel_server_urls.append(pbi.wheel_server_url)
    else:
        if ctx.wheel_server_url:
            # local wheel server
            wheel_server_urls.append(ctx.wheel_server_url)
        if cache_wheel_server_url:
            # put cache after local server so we always check local server first
            wheel_server_urls.append(cache_wheel_server_url)
        if not wheel_server_urls:
            raise ValueError("no wheel server urls configured")
    return wheel_server_urls


@metrics.timeit(description="resolve wheel")
def resolve_prebuilt_wheel(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    wheel_server_urls: list[str],
    req_type: requirements_file.RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to wheel and its version."
    excs: list[Exception] = []
    for url in wheel_server_urls:
        try:
            wheel_url, resolved_version = resolver.resolve(
                ctx=ctx,
                req=req,
                sdist_server_url=url,
                include_sdists=False,
                include_wheels=True,
                req_type=req_type,
                # pre-built wheels must match platform
                ignore_platform=False,
            )
        except Exception as e:
            excs.append(e)
        else:
            if wheel_url and resolved_version:
                return (wheel_url, resolved_version)
            else:
                excs.append(
                    ValueError(
                        f"no result for {url}: {wheel_url=}, {resolved_version=}"
                    )
                )
    raise ExceptionGroup(
        f"Could not find a prebuilt wheel for {req} on {' or '.join(wheel_server_urls)}",
        excs,
    )



import csv
import json
import logging
import pathlib
import sys
from enum import Enum
from typing import TextIO

import click
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import constraints, context, overrides, resolver
from fromager.commands.list_versions import DistributionType

logger = logging.getLogger(__name__)


class OutputFormat(Enum):
    """Output format for find-updates command"""

    REQUIREMENTS = "requirements"
    JSON = "json"
    CSV = "csv"


@click.command()
@click.option(
    "--format",
    "output_format",
    type=click.Choice(OutputFormat, case_sensitive=False),
    default=OutputFormat.REQUIREMENTS.value,
    help="Output format (requirements: name==version, json: JSON array, csv: CSV with name,version columns)",
)
@click.option(
    "--distribution-type",
    type=click.Choice(DistributionType, case_sensitive=False),
    default=DistributionType.DEFAULT.value,
    help="Distribution type to include in version lookup (default: use package settings, sdist: source only, wheel: wheels only, both: include both sdists and wheels)",
)
@click.option(
    "--sdist-server-url",
    default=resolver.PYPI_SERVER_URL,
    help="URL to the Python package index to use for version lookup",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(path_type=pathlib.Path),
    help="Write output to file instead of stdout",
)
@click.argument(
    "constraints_file", type=click.Path(exists=True, path_type=pathlib.Path)
)
@click.pass_obj
def find_updates(
    wkctx: context.WorkContext,
    constraints_file: pathlib.Path,
    output_format: str,
    distribution_type: str,
    sdist_server_url: str,
    output: pathlib.Path | None,
) -> None:
    """Find available updates for packages with specific version constraints.

    This command reads a constraints file and for each package that has a
    supported constraint (==, <, <=, ~=, !=, ===), it lists newer versions available.
    Constraints with lower bounds (>=) are ignored as they don't pin to specific versions.

    The CONSTRAINTS_FILE should be a pip-style constraints file with entries like:
    - "package_name==1.0.0" (will look for versions > 1.0.0)
    - "package_name<2.0.0" (will look for versions >= 2.0.0)
    - "package_name<=1.5.0" (will look for versions > 1.5.0)
    - "package_name~=1.4.0" (will look for versions outside compatible range)
    - "package_name!=1.0.0" (will look for versions other than 1.0.0)
    - "package_name===1.0.0" (will look for versions other than exactly 1.0.0)
    - "package_name>=1.0.0" (ignored - not a specific constraint)

    Distribution types:
    - "default": Use package settings for include_sdists/include_wheels
    - "sdist": Only include source distributions
    - "wheel": Only include wheels
    - "both": Include both source distributions and wheels

    Output formats:
    - "requirements": List as requirement specifiers (package==version)
    - "json": JSON array of objects with "name" and "version" fields
    - "csv": CSV format with "name" and "version" columns

    Use --output to write results to a file instead of stdout.
    """
    # Load the constraints file
    constraint_loader = constraints.Constraints()
    constraint_loader.load_constraints_file(constraints_file)

    updates_found: list[dict[str, str]] = []

    # Process each constraint
    for package_name in constraint_loader:
        constraint = constraint_loader.get_constraint(package_name)
        if not constraint:
            continue

        # Check if this is a constraint we can check for updates (==, <, <=)
        constraint_version = _get_constraint_version(constraint)
        if not constraint_version:
            logger.debug(f"Skipping {package_name}: not a supported constraint type")
            continue

        logger.info(f"Looking for updates to {package_name} (constraint: {constraint})")

        # Find available versions for this package
        try:
            newer_versions = _find_newer_versions(
                wkctx,
                constraint,
                constraint_version,
                distribution_type,
                sdist_server_url,
            )

            if newer_versions:
                logger.info(
                    f"Found {len(newer_versions)} newer version(s) for {package_name}"
                )
                for version in newer_versions:
                    updates_found.append(
                        {"name": package_name, "version": str(version)}
                    )
            else:
                logger.debug(f"No newer versions found for {package_name}")

        except Exception as e:
            logger.warning(f"Could not check versions for {package_name}: {e}")

    # Output results in requested format
    if not updates_found:
        return

    # Determine output destination
    output_file: TextIO
    if output:
        output_file = output.open("w")
    else:
        output_file = sys.stdout

    try:
        output_fmt = OutputFormat(output_format)
        if output_fmt == OutputFormat.REQUIREMENTS:
            for update in updates_found:
                print(f"{update['name']}=={update['version']}", file=output_file)
        elif output_fmt == OutputFormat.JSON:
            json.dump(updates_found, output_file, indent=2)
            print(file=output_file)  # Add newline for better output
        elif output_fmt == OutputFormat.CSV:
            writer = csv.writer(output_file)
            writer.writerow(["name", "version"])
            for update in updates_found:
                writer.writerow([update["name"], update["version"]])
    finally:
        if output:
            output_file.close()


CONSTRAINT_OPERATORS = ("==", "<", "<=", "~=", "!=", "===")


def _get_constraint_version(constraint: Requirement) -> Version | None:
    """Extract version from supported constraint types (==, <, <=, ~=, !=, ===), return None if not supported."""
    if not constraint.specifier:
        return None

    # Look for supported constraint operators
    supported_specs = [
        spec for spec in constraint.specifier if spec.operator in CONSTRAINT_OPERATORS
    ]
    if len(supported_specs) != 1:
        return None

    # Parse the version from the constraint specifier
    try:
        return Version(supported_specs[0].version)
    except Exception:
        return None


def _find_newer_versions(
    wkctx: context.WorkContext,
    constraint: Requirement,
    constraint_version: Version,
    distribution_type: str,
    sdist_server_url: str,
) -> list[Version]:
    """Find versions that would be updates beyond the constraint for a package."""
    # Get package build info to determine distribution preferences
    pbi = wkctx.package_build_info(constraint)
    override_sdist_server_url = pbi.resolver_sdist_server_url(sdist_server_url)

    # Determine include flags based on distribution type
    dist_type = DistributionType(distribution_type)
    match dist_type:
        case DistributionType.SDIST:
            include_sdists = True
            include_wheels = False
        case DistributionType.WHEEL:
            include_sdists = False
            include_wheels = True
        case DistributionType.BOTH:
            include_sdists = True
            include_wheels = True
        case _:  # DEFAULT
            # Use package settings defaults
            package_settings = wkctx.settings.package_setting(constraint.name)
            include_sdists = package_settings.resolver_dist.include_sdists
            include_wheels = package_settings.resolver_dist.include_wheels

    # Get resolver provider
    provider = overrides.find_and_invoke(
        constraint.name,
        "get_resolver_provider",
        resolver.default_resolver_provider,
        ctx=wkctx,
        req=constraint,
        include_sdists=include_sdists,
        include_wheels=include_wheels,
        sdist_server_url=override_sdist_server_url,
    )

    # Create a requirement without version constraints to get all available versions
    unconstrained_req = Requirement(constraint.name)

    # Get all available candidates
    candidates = list(
        provider.find_matches(
            identifier=constraint.name,
            requirements={constraint.name: [unconstrained_req]},
            incompatibilities={constraint.name: []},
        )
    )

    if not candidates:
        return []

    # Extract versions and filter based on constraint type
    all_versions = sorted(set(candidate.version for candidate in candidates))

    # Get the constraint operator to determine what "update" means
    constraint_spec = next(
        spec for spec in constraint.specifier if spec.operator in CONSTRAINT_OPERATORS
    )
    if constraint_spec.operator in ("==", "==="):
        # For equality constraints, find versions newer than the pinned version
        newer_versions = [v for v in all_versions if v > constraint_version]
    elif constraint_spec.operator in ("<", "<=", "~=", "!="):
        # For upper bound constraints, find versions that exceed the constraint
        # (these would be versions that violate the current constraint)
        newer_versions = [v for v in all_versions if v >= constraint_version]
    else:
        newer_versions = []

    return newer_versions



from __future__ import annotations

import inspect
import json
import logging
import pathlib
import shutil
import tarfile
import typing
import zipfile

import resolvelib
from packaging.requirements import Requirement
from packaging.utils import (
    parse_sdist_filename,
)
from packaging.version import Version
from requests.exceptions import ChunkedEncodingError, ConnectionError
from urllib3.exceptions import IncompleteRead, ProtocolError

from . import (
    build_environment,
    dependencies,
    external_commands,
    gitutils,
    metrics,
    overrides,
    packagesettings,
    pyproject,
    requirements_file,
    resolver,
    tarballs,
    vendor_rust,
)
from .http_retry import RETRYABLE_EXCEPTIONS, retry_on_exception
from .request_session import session
from .requirements_file import RequirementType

if typing.TYPE_CHECKING:
    from . import build_environment, context

logger = logging.getLogger(__name__)


def get_source_type(ctx: context.WorkContext, req: Requirement) -> str:
    source_type = requirements_file.SourceType.SDIST
    if req.url:
        return requirements_file.SourceType.GIT
    pbi = ctx.package_build_info(req)
    if (
        overrides.find_override_method(req.name, "download_source")
        or overrides.find_override_method(req.name, "resolve_source")
        or overrides.find_override_method(req.name, "get_resolver_provider")
        or pbi.download_source_url(resolve_template=False)
    ):
        source_type = requirements_file.SourceType.OVERRIDE
    return str(source_type)


@metrics.timeit(description="download source")
def download_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    download_url: str,
) -> pathlib.Path:
    logger.info(f"downloading source for {req}")
    download_path = pathlib.Path(download_url)
    if req.url and download_path.exists():
        logger.info(
            "source is already downloaded to %s by cloning %s, ignoring any plugins",
            download_url,
            req.url,
        )
        return download_path
    elif req.url:
        download_path = ctx.work_dir / f"{req.name}-{version}" / f"{req.name}-{version}"
        download_path.mkdir(parents=True, exist_ok=True)
        download_git_source(
            ctx=ctx,
            req=req,
            url_to_clone=req.url,
            destination_dir=download_path,
        )
        return download_path

    source_path = overrides.find_and_invoke(
        req.name,
        "download_source",
        default_download_source,
        ctx=ctx,
        req=req,
        version=version,
        download_url=download_url,
        sdists_downloads_dir=ctx.sdists_downloads,
    )

    if not isinstance(source_path, pathlib.Path):
        raise ValueError(
            f"expected a Path back to downloaded source. got {source_path}"
        )
    return source_path


@metrics.timeit(description="resolve source")
def resolve_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    req_type: RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to source and its version."

    constraint = ctx.constraints.get_constraint(req.name)
    logger.debug(
        f"resolving requirement {req} using {sdist_server_url} with constraint {constraint}"
    )

    try:
        resolver_results = overrides.find_and_invoke(
            req.name,
            "resolve_source",
            default_resolve_source,
            ctx=ctx,
            req=req,
            sdist_server_url=sdist_server_url,
            req_type=req_type,
        )
    except (
        resolvelib.InconsistentCandidate,
        resolvelib.RequirementsConflicted,
        resolvelib.ResolutionImpossible,
    ) as err:
        logger.debug(f"could not resolve {req} with {constraint}: {err}")
        raise

    if len(resolver_results) == 2:
        url, version = resolver_results
    else:
        raise ValueError(
            f"do not know how to unpack {resolver_results}, expected 2 members"
        )

    if not isinstance(version, Version):
        raise ValueError(f"expected 2nd member to be of type Version, got {version}")

    return str(url), version


def default_resolve_source(
    ctx: context.WorkContext,
    req: Requirement,
    sdist_server_url: str,
    req_type: RequirementType | None = None,
) -> tuple[str, Version]:
    "Return URL to source and its version."

    pbi = ctx.package_build_info(req)
    override_sdist_server_url = pbi.resolver_sdist_server_url(sdist_server_url)

    url, version = resolver.resolve(
        ctx=ctx,
        req=req,
        sdist_server_url=override_sdist_server_url,
        include_sdists=pbi.resolver_include_sdists,
        include_wheels=pbi.resolver_include_wheels,
        req_type=req_type,
        ignore_platform=pbi.resolver_ignore_platform,
    )
    return url, version


def default_download_source(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    download_url: str,
    sdists_downloads_dir: pathlib.Path,
) -> pathlib.Path:
    "Download the requirement and return the name of the output path."
    pbi = ctx.package_build_info(req)
    destination_filename = pbi.download_source_destination_filename(version=version)
    url = pbi.download_source_url(version=version, default=download_url)
    source_filename = _download_source_check(
        req=req,
        destination_dir=sdists_downloads_dir,
        url=url,
        destination_filename=destination_filename,
    )

    logger.debug(f"have source for {req} version {version} in {source_filename}")
    return source_filename


def download_git_source(
    ctx: context.WorkContext,
    req: Requirement,
    url_to_clone: str,
    destination_dir: pathlib.Path,
    ref: str | None = None,
) -> None:
    if url_to_clone.startswith("git+"):
        url_to_clone = url_to_clone[len("git+") :]

    logger.info(f"cloning source from {url_to_clone}@{ref} to {destination_dir}")
    # Get git options from package settings
    pbi = ctx.package_build_info(req)
    git_opts = pbi.git_options

    # Configure submodules based on package settings
    submodules: bool | list[str] = False
    if git_opts.submodule_paths:
        # If specific paths are configured, use those
        submodules = git_opts.submodule_paths
    elif git_opts.submodules:
        # If general submodule support is enabled, clone all submodules
        submodules = True

    gitutils.git_clone(
        ctx=ctx,
        req=req,
        output_dir=destination_dir,
        repo_url=url_to_clone,
        submodules=submodules,
        ref=ref,
    )


# Helper method to check whether .zip /.tar / .tgz is able to extract and check its content.
# It will throw exception if any other file is encountered. Eg: index.html
def _download_source_check(
    req: Requirement,
    destination_dir: pathlib.Path,
    url: str,
    destination_filename: str | None = None,
) -> str:
    source_filename = download_url(
        req=req,
        destination_dir=destination_dir,
        url=url,
        destination_filename=destination_filename,
    )
    if source_filename.suffix == ".zip":
        with zipfile.ZipFile(source_filename) as zip_file:
            source_file_contents = zip_file.namelist()
            if not source_file_contents:
                raise zipfile.BadZipFile(
                    f"Empty zip file encountered: {source_filename}"
                )
    elif (
        source_filename.suffix == ".tgz"
        or source_filename.suffix == ".gz"
        or str(source_filename).endswith(".tar.gz")
    ):
        with tarfile.open(source_filename) as tar:
            if not tar.next():
                raise tarfile.TarError(f"Empty tar file encountered: {source_filename}")
    else:
        raise ValueError(
            f"The source file encountered is not a zip or tar file: {source_filename}"
        )
    return source_filename


def download_url(
    *,
    req: Requirement,
    destination_dir: pathlib.Path,
    url: str,
    destination_filename: str | None = None,
) -> pathlib.Path:
    basename = (
        destination_filename
        if destination_filename
        else resolver.extract_filename_from_url(url)
    )
    outfile = pathlib.Path(destination_dir) / basename
    logger.debug(
        "looking for %s %s",
        outfile,
        "(exists)" if outfile.exists() else "(not there)",
    )
    if outfile.exists():
        logger.debug("already have %s", outfile)
        return outfile

    # Create a temporary file to avoid partial downloads
    temp_file = outfile.with_suffix(outfile.suffix + ".tmp")

    def _download_with_retry():
        """Internal function that performs the actual download with retry logic."""
        logger.debug(f"reading from {url}")
        try:
            with session.get(url, stream=True) as r:
                r.raise_for_status()
                with open(temp_file, "wb") as f:
                    logger.debug("writing to %s", temp_file)
                    # Use smaller chunk size for better error recovery
                    for chunk in r.iter_content(chunk_size=64 * 1024):
                        if chunk:  # Filter out keep-alive chunks
                            f.write(chunk)

            # Only move to final location if download completed successfully
            temp_file.rename(outfile)
            logger.info("saved %s", outfile)
            return outfile

        except (
            ChunkedEncodingError,
            IncompleteRead,
            ProtocolError,
            ConnectionError,
        ) as e:
            # Clean up partial file on failure
            if temp_file.exists():
                temp_file.unlink()
            logger.warning(f"Download failed for {url}: {e}")
            raise
        except Exception:
            # Clean up partial file on any other failure
            if temp_file.exists():
                temp_file.unlink()
            raise

    # Apply retry logic specifically for download operations
    @retry_on_exception(
        exceptions=RETRYABLE_EXCEPTIONS,
        max_attempts=5,
        backoff_factor=1.5,
        max_backoff=120.0,
    )
    def download_with_retry():
        return _download_with_retry()

    try:
        return download_with_retry()
    except Exception:
        # Ensure temp file is cleaned up if it still exists
        if temp_file.exists():
            temp_file.unlink()
        raise


def _takes_arg(f: typing.Callable, arg_name: str) -> bool:
    sig = inspect.signature(f)
    return arg_name in sig.parameters


def unpack_source(
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    source_filename: pathlib.Path,
) -> tuple[pathlib.Path, bool]:
    # sdist names are less standardized and the names of the directories they
    # contain are also not very standard. Force the names into a predictable
    # form based on the override module name for the requirement.
    req_name = overrides.pkgname_to_override_module(req.name)
    expected_name = f"{req_name}-{version}"

    # The unpack_dir is a parent dir where we put temporary outputs during the
    # build process, including the unpacked source in a subdirectory.
    unpack_dir = ctx.work_dir / expected_name
    if unpack_dir.exists():
        if ctx.cleanup:
            logger.debug("cleaning up %s", unpack_dir)
            shutil.rmtree(unpack_dir)
        else:
            logger.info("reusing %s", unpack_dir)
            return (unpack_dir / unpack_dir.name, False)

    # sdists might be tarballs or zip files.
    logger.debug("unpacking %s to %s", source_filename, unpack_dir)
    if str(source_filename).endswith(".tar.gz"):
        with tarfile.open(source_filename, "r") as t:
            if _takes_arg(t.extractall, "filter"):
                t.extractall(unpack_dir, filter="data")
            else:
                logger.debug('unpacking without filter="data"')
                t.extractall(unpack_dir)
    elif str(source_filename).endswith(".zip"):
        with zipfile.ZipFile(source_filename) as zf:
            zf.extractall(path=unpack_dir)
    else:
        raise ValueError(f"Do not know how to unpack source archive {source_filename}")

    # We create a unique directory based on the requirement name, but that may
    # not be the same name as the root directory of the content in the sdist
    # (due to case, punctuation, etc.), so after we unpack it look for what was
    # created and ensure the extracted directory matches the override module
    # name and version of the requirement.
    unpacked_root_dir = next(iter(unpack_dir.glob("*")))
    if unpacked_root_dir.name != expected_name:
        desired_name = unpacked_root_dir.parent / expected_name
        try:
            shutil.move(
                str(unpacked_root_dir),
                str(desired_name),
            )
        except Exception as err:
            raise Exception(
                f"Could not rename {unpacked_root_dir.name} to {desired_name}: {err}"
            ) from err
        unpacked_root_dir = desired_name

    return (unpacked_root_dir, True)


def patch_source(
    ctx: context.WorkContext,
    source_root_dir: pathlib.Path,
    req: Requirement,
    version: Version,
) -> None:
    pbi = ctx.package_build_info(req)
    patch_count = 0

    for p in pbi.get_patches(version):
        _apply_patch(req, p, source_root_dir)
        patch_count += 1

    logger.debug("applied %d patches", patch_count)
    # If no patch has been applied, call warn for old patch
    patchmap = pbi.get_all_patches()
    if not patch_count and patchmap:
        for patchversion in sorted(patchmap):
            logger.warning(
                f"patch {patchversion} exists but will not be applied for version {version}"
            )


def _apply_patch(req: Requirement, patch: pathlib.Path, source_root_dir: pathlib.Path):
    logger.info("applying patch file %s to %s", patch, source_root_dir)
    with open(patch, "r") as f:
        external_commands.run(["patch", "-p1"], stdin=f, cwd=source_root_dir)


def write_build_meta(
    unpack_dir: pathlib.Path,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> pathlib.Path:
    meta_file = unpack_dir / "build-meta.json"
    with open(meta_file, "w") as f:
        json.dump(
            {
                "req": str(req),
                "source-filename": str(source_filename),
                "version": str(version),
            },
            f,
        )
    logger.debug("wrote build metadata to %s", meta_file)
    return meta_file


def read_build_meta(unpack_dir: pathlib.Path) -> dict:
    meta_file = unpack_dir / "build-meta.json"
    with open(meta_file, "r") as f:
        return json.load(f)


@metrics.timeit(description="prepare source")
def prepare_source(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> pathlib.Path:
    if req.url:
        logger.info(
            "preparing source cloned from %s into %s, ignoring any plugins",
            req.url,
            source_filename,
        )
        source_root_dir = pathlib.Path(source_filename)
        prepare_new_source(
            ctx=ctx,
            req=req,
            source_root_dir=source_root_dir,
            version=version,
        )
    else:
        logger.info(f"preparing source for {req} from {source_filename}")
        prepare_source_details = overrides.find_and_invoke(
            req.name,
            "prepare_source",
            default_prepare_source,
            ctx=ctx,
            req=req,
            source_filename=source_filename,
            version=version,
        )
        if not isinstance(prepare_source_details, tuple):
            source_root_dir = prepare_source_details
        elif len(prepare_source_details) == 2:
            source_root_dir, _ = prepare_source_details
        else:
            raise ValueError(
                f"do not know how to unpack {prepare_source_details}, expected 1 or 2 members"
            )
    write_build_meta(source_root_dir.parent, req, source_filename, version)
    if source_root_dir is not None:
        logger.info(f"prepared source for {req} at {source_root_dir}")
    return source_root_dir


def default_prepare_source(
    ctx: context.WorkContext,
    req: Requirement,
    source_filename: pathlib.Path,
    version: Version,
) -> tuple[pathlib.Path, bool]:
    """Unpack and modify sdist sources

    Calls :func:`~fromager.sources.prepare_new_source` by default.
    """
    source_root_dir, is_new = unpack_source(
        ctx=ctx,
        req=req,
        version=version,
        source_filename=source_filename,
    )
    if is_new:
        prepare_new_source(
            ctx=ctx,
            req=req,
            source_root_dir=source_root_dir,
            version=version,
        )
    return source_root_dir, is_new


def prepare_new_source(
    ctx: context.WorkContext,
    req: Requirement,
    source_root_dir: pathlib.Path,
    version: Version,
) -> None:
    """Default steps for new sources

    - patch sources
    - apply project overrides from settings
    - vendor Rust dependencies

    :func:`~default_prepare_source` runs this function when the sources are new.
    """
    patch_source(ctx, source_root_dir, req, version)
    pyproject.apply_project_override(
        ctx=ctx,
        req=req,
        sdist_root_dir=source_root_dir,
    )
    vendor_rust.vendor_rust(req, source_root_dir)


@metrics.timeit(description="build sdist")
def build_sdist(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    """Build source distribution"""
    pbi = ctx.package_build_info(req)
    build_dir = pbi.build_dir(sdist_root_dir)

    logger.info(f"building {ctx.variant} source distribution for {req} in {build_dir}")
    extra_environ = packagesettings.get_extra_environ(
        ctx=ctx,
        req=req,
        version=version,
        sdist_root_dir=sdist_root_dir,
        build_env=build_env,
    )
    if req.url:
        # The default approach to making an sdist is to make a tarball from the
        # source directory, since most of the time we got the source directory
        # by unpacking an existing sdist. When we know we cloned a git repo to
        # get the source tree, we can be very sure that creating a tarball will
        # NOT produce a valid sdist, so we can use the PEP-517 approach
        # instead.
        logger.info("using PEP-517 sdist build, ignoring any plugins")
        sdist_file = pep517_build_sdist(
            ctx=ctx,
            extra_environ=extra_environ,
            req=req,
            sdist_root_dir=sdist_root_dir,
            version=version,
            build_env=build_env,
        )
    else:
        sdist_file = overrides.find_and_invoke(
            req.name,
            "build_sdist",
            default_build_sdist,
            ctx=ctx,
            extra_environ=extra_environ,
            req=req,
            version=version,
            sdist_root_dir=sdist_root_dir,
            build_dir=build_dir,
            build_env=build_env,
        )
    logger.info(f"built source distribution {sdist_file}")

    # validate location and file name
    if not sdist_file.is_file():
        raise FileNotFoundError(sdist_file)
    sdist_file = sdist_file.resolve()
    if sdist_file.parent != ctx.sdists_builds:
        raise ValueError(
            f"{sdist_file!r} is not in sdists build directory {ctx.sdists_builds!r}"
        )
    validate_sdist_filename(req=req, version=version, sdist_file=sdist_file)

    return sdist_file


def default_build_sdist(
    ctx: context.WorkContext,
    extra_environ: dict,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_env: build_environment.BuildEnvironment,
    build_dir: pathlib.Path,
) -> pathlib.Path:
    # It seems like the "correct" way to do this would be to run the
    # PEP 517 API in the source tree we have modified. However, quite
    # a few packages assume their source distribution is being built
    # from a source code repository checkout and those throw an error
    # when we use the interface to try to rebuild the sdist. Since we
    # know what we have is an exploded tarball, we just tar it back
    # up.
    #
    # For cases where the PEP 517 approach works, use
    # pep517_build_sdist().
    sdist_filename = ctx.sdists_builds / f"{req.name}-{version}.tar.gz"
    if sdist_filename.exists():
        sdist_filename.unlink()
    ensure_pkg_info(
        ctx=ctx,
        req=req,
        version=version,
        sdist_root_dir=sdist_root_dir,
        build_dir=build_dir,
    )
    # The format argument is specified based on
    # https://peps.python.org/pep-0517/#build-sdist.
    with tarfile.open(sdist_filename, "x:gz", format=tarfile.PAX_FORMAT) as sdist:
        tarballs.tar_reproducible(
            tar=sdist,
            basedir=build_dir,
            prefix=build_dir.parent,
        )
    return sdist_filename


def pep517_build_sdist(
    ctx: context.WorkContext,
    extra_environ: dict,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    """Use the PEP 517 API to build a source distribution from a modified source tree."""
    pbi = ctx.package_build_info(req)
    build_dir = pbi.build_dir(sdist_root_dir)
    hook_caller = dependencies.get_build_backend_hook_caller(
        ctx=ctx,
        req=req,
        build_dir=build_dir,
        override_environ=extra_environ,
        build_env=build_env,
    )
    sdist_filename = hook_caller.build_sdist(
        ctx.sdists_builds,
        config_settings=pbi.config_settings,
    )
    return ctx.sdists_builds / sdist_filename


PKG_INFO_CONTENT = """\
Metadata-Version: 1.0
Name: {name}
Version: {version}
Summary: {summary}
"""


def ensure_pkg_info(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version,
    sdist_root_dir: pathlib.Path,
    build_dir: pathlib.Path | None = None,
) -> bool:
    """Ensure that sdist has a PKG-INFO file

    Returns True if PKG-INFO was presence, False if file was missing. The
    function also updates build_dir if package has a non-standard build
    directory. Every sdist must have a PKG-INFO file in the first directory.
    The additional PKG-INFO file in build_dir is required for projects
    with non-standard layout and setuptools-scm.
    """
    had_pkg_info = True
    directories = [sdist_root_dir]
    if build_dir is not None and build_dir != sdist_root_dir:
        directories.append(build_dir)
    for directory in directories:
        pkg_info_file = directory / "PKG-INFO"
        if not pkg_info_file.is_file():
            logger.warning(
                f"PKG-INFO file is missing from {directory}, creating stub file"
            )
            pkg_info_file.write_text(
                PKG_INFO_CONTENT.format(
                    name=req.name,
                    version=str(version),
                    summary=dependencies.STUB_PKG_INFO_SUMMARY,
                )
            )
            had_pkg_info = False
    return had_pkg_info


def validate_sdist_filename(
    req: Requirement,
    version: Version,
    sdist_file: pathlib.Path,
) -> None:
    """Check that sdist matches requirement name and version"""
    sdist_name, sdist_version = parse_sdist_filename(sdist_file.name)
    dependencies.validate_dist_name_version(
        req=req,
        version=version,
        what=sdist_file.name,
        dist_name=sdist_name,
        dist_version=sdist_version,
    )



import dataclasses

import pytest
from packaging.utils import canonicalize_name
from packaging.version import Version

from fromager.dependency_graph import DependencyNode


def mknode(name: str, version: str = "1.0", **kwargs) -> DependencyNode:
    return DependencyNode(canonicalize_name(name), Version(version), **kwargs)


def test_dependencynode_compare() -> None:
    a_10 = mknode("a", "1.0")
    a_20 = mknode("a", "2.0")
    b = mknode("b", "1.0")
    assert a_10 == a_10
    assert not a_10 == a_20
    assert a_10 != a_20
    assert a_10 != b
    assert a_10 == mknode("a", "1.0")
    assert a_10 < a_20
    assert a_10 <= a_10
    assert a_10 >= a_10
    assert b > a_10
    assert b > a_20


def test_dependencynode_hash() -> None:
    a_10 = mknode("a", "1.0")
    a_20 = mknode("a", "2.0")
    b = mknode("b", "1.0")
    s = {a_10, a_10, a_20}
    assert s == {a_10, a_20}
    assert a_10 in s
    assert b not in s

    s = {mknode("a"), mknode("a")}
    assert len(s) == 1
    assert s == {mknode("a")}


def test_dependencynode_dataclass():
    a = mknode("a", "1.0")
    assert a.canonicalized_name == "a"
    assert a.version == Version("1.0")
    assert a.key == "a==1.0"
    assert (
        repr(a)
        == "DependencyNode(canonicalized_name='a', version=<Version('1.0')>, download_url='', pre_built=False)"
    )
    with pytest.raises(dataclasses.FrozenInstanceError):
        a.version = Version("2.0")
    with pytest.raises((TypeError, AttributeError)):
        a.new_attribute = None

    root = DependencyNode.construct_root_node()
    assert root.canonicalized_name == ""
    assert root.version == Version("0.0")
    assert root.key == ""



import csv
import json
import pathlib
import sys

import click
import rich
from packaging.version import Version
from rich.table import Table

from fromager import clickext, context
from fromager.packagesettings import PatchMap


@click.command()
@click.option(
    "--details",
    is_flag=True,
    default=False,
    help="Show more details about the overrides.",
)
@click.option(
    "--format",
    "output_format",
    type=click.Choice(["table", "csv", "json"], case_sensitive=False),
    default="table",
    help="Output format for detailed view (requires --details, default: table)",
)
@click.option(
    "-o",
    "--output",
    type=clickext.ClickPath(),
    help="Output file to create (requires --details, default: stdout)",
)
@click.pass_obj
def list_overrides(
    wkctx: context.WorkContext,
    details: bool,
    output_format: str,
    output: pathlib.Path | None,
) -> None:
    """List all of the packages with overrides in the current configuration."""
    # Warn if format/output options are used without --details
    if not details:
        if output_format != "table":
            click.echo(
                "Warning: --format option is ignored when --details is not used",
                err=True,
            )
        if output is not None:
            click.echo(
                "Warning: --output option is ignored when --details is not used",
                err=True,
            )

    overridden_packages = sorted(wkctx.settings.list_overrides())
    if not details:
        for name in overridden_packages:
            print(name)
        return

    # Collect data for export
    variants = sorted(wkctx.settings.all_variants())
    variant_names = [str(v) for v in variants]
    export_data = []

    for name in overridden_packages:
        pbi = wkctx.settings.package_build_info(name)
        ps = wkctx.settings.package_setting(name)

        plugin_hooks: list[str] = []
        if pbi.plugin:
            for hook in [
                # from hooks.py
                "post_build",
                "post_bootstrap",
                "prebuilt_wheel",
                # from overrides.py, found by searching for find_override_method
                "download_source",
                "resolve_source",
                "get_resolver_provider",
                "prepare_source",
                "build_sdist",
                "build_wheel",
                "get_build_requirements",
                "get_build_sdist_requirements",
                "get_build_wheel_requirements",
                "expected_source_archive_name",
                "expected_source_directory_name",
                "add_extra_metadata_to_wheels",
            ]:
                if hasattr(pbi.plugin, hook):
                    plugin_hooks.append(hook)
        plugin_hooks_str = ", ".join(plugin_hooks)

        variant_info: dict[str, str] = {}
        for v in variants:
            v_info = ps.variants.get(v)
            if v_info:
                if v_info.pre_built:
                    variant_info[str(v)] = "pre-built"
                else:
                    variant_info[str(v)] = "yes"
            else:
                variant_info[str(v)] = ""

        all_patches: PatchMap = pbi.get_all_patches()
        global_patches: list[pathlib.Path] = all_patches.get(None, [])
        num_global_patches: int = len(global_patches)

        all_pkg_versions: list[Version] = sorted(
            [v for v in all_patches.keys() if v is not None]
        )

        if not all_pkg_versions:
            # This package has overrides, but none are version-specific.
            patches_str = str(num_global_patches) if num_global_patches else ""
            row_data = {
                "package": name,
                "version": "",
                "patches": patches_str,
                "plugin_hooks": plugin_hooks_str,
            }
            # Add variant information
            row_data.update(variant_info)
            export_data.append(row_data)
        else:
            # This package has version-specific overrides.
            for version in all_pkg_versions:
                version_patches: list[pathlib.Path] = all_patches.get(version, [])
                total_patches: int = num_global_patches + len(version_patches)
                patches_str = str(total_patches) if total_patches else ""

                row_data = {
                    "package": name,
                    "version": str(version),
                    "patches": patches_str,
                    "plugin_hooks": plugin_hooks_str,
                }
                # Add variant information
                row_data.update(variant_info)
                export_data.append(row_data)

    # Handle different output formats
    match output_format:
        case "json":
            _export_json(export_data, output)
        case "csv":
            _export_csv(export_data, variant_names, output)
        case "table":
            _export_table(export_data, variant_names)
        case _:
            raise ValueError(f"Invalid output format: {output_format}")


def _export_json(data: list[dict], output: pathlib.Path | None) -> None:
    """Export data as JSON."""
    if output:
        with open(output, "w") as outfile:
            json.dump(data, outfile, indent=2)
    else:
        json.dump(data, sys.stdout, indent=2)


def _export_csv(
    data: list[dict], variants: list[str], output: pathlib.Path | None
) -> None:
    """Export data as CSV."""
    # Define field names in the order we want them
    fieldnames = ["package", "version", "patches"] + variants + ["plugin_hooks"]

    if output:
        with open(output, "w", newline="") as outfile:
            writer = csv.DictWriter(
                outfile, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
            )
            writer.writeheader()
            writer.writerows(data)
    else:
        writer = csv.DictWriter(
            sys.stdout, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
        )
        writer.writeheader()
        writer.writerows(data)


def _export_table(data: list[dict], variants: list[str]) -> None:
    """Export data as Rich table (original behavior)."""
    table = Table(title="Package Overrides")
    table.add_column("Package", justify="left", no_wrap=True)
    table.add_column("Version", justify="left", no_wrap=True)
    table.add_column("Patches", justify="left", no_wrap=True)

    for v in variants:
        table.add_column(v, justify="left", no_wrap=True)

    table.add_column("Plugin", justify="left")

    # Define column keys in the same order as CSV exporter
    column_keys = ["package", "version", "patches"] + variants + ["plugin_hooks"]

    for row_data in data:
        row = [row_data.get(key, "") for key in column_keys]
        table.add_row(*row)

    rich.get_console().print(table)



import textwrap
from unittest.mock import Mock, patch

from packaging.requirements import Requirement
from packaging.version import Version

from fromager import build_environment
from fromager.context import WorkContext
from fromager.requirements_file import RequirementType


@patch("fromager.resolver.resolve")
def test_missing_dependency_format(
    resolve_dist: Mock,
    tmp_context: WorkContext,
):
    resolutions = {
        "flit_core": "3.9.0",
        "setuptools": "69.5.1",
    }
    resolve_dist.side_effect = lambda ctx, req, sdist_server_url, req_type: (
        "",
        Version(resolutions[req.name]),
    )

    req = Requirement("setuptools>=40.8.0")
    other_reqs = [
        Requirement("flit_core"),
        req,
    ]
    ex = build_environment.MissingDependency(
        tmp_context, RequirementType.BUILD_BACKEND, req, other_reqs
    )
    s = str(ex)
    # Ensure we report the thing we're actually missing
    assert "Failed to install build-backend dependency setuptools>=40.8.0. " in s
    # Ensure we report what version we expected of that thing
    assert "setuptools>=40.8.0 -> 69.5.1" in s
    # Ensure we report what version we expect of all of the other dependencies
    assert "flit_core -> 3.9.0" in s


def test_missing_dependency_pattern():
    msg = textwrap.dedent("""
        DEBUG uv 0.8.4
        DEBUG Searching for default Python interpreter in virtual environments
        DEBUG Found `cpython-3.13.5-linux-x86_64-gnu` at `.../.venv/bin/python3` (active virtual environment)
        DEBUG Using Python 3.13.5 environment at: .venv
        DEBUG Acquired lock for `.venv`
        DEBUG At least one requirement is not satisfied: fromager==1.0
        DEBUG Using request timeout of 30s
        DEBUG Solving with installed Python version: 3.13.5
        DEBUG Solving with target Python version: >=3.13.5
        DEBUG Adding direct dependency: fromager>=1.0, <1.0+
        DEBUG Found fresh response for: https://pypi.org/simple/fromager/
        DEBUG Searching for a compatible version of fromager (>=1.0, <1.0+)
        DEBUG No compatible version found for: fromager
        x No solution found when resolving dependencies:
        ╰─▶ Because there is no version of fromager==1.0 and you require fromager==1.0, we can conclude that your requirements are unsatisfiable.
        """)
    match = build_environment._uv_missing_dependency_pattern.search(msg)
    assert match is not None


def test_missing_dependency_pattern_resolution_impossible():
    msg = textwrap.dedent("""
        DEBUG uv 0.8.4
        DEBUG Searching for default Python interpreter in virtual environments
        DEBUG Found `cpython-3.13.5-linux-x86_64-gnu` at `.../.venv/bin/python3` (active virtual environment)
        DEBUG Using Python 3.13.5 environment at: .venv
        DEBUG Acquired lock for `.venv`
        DEBUG At least one requirement is not satisfied: fromager==2.0
        DEBUG Using request timeout of 30s
        DEBUG Solving with installed Python version: 3.13.5
        DEBUG Solving with target Python version: >=3.13.5
        DEBUG Adding direct dependency: fromager>=1.0, <1.0+
        DEBUG Adding direct dependency: fromager>=2.0, <2.0+
        DEBUG Found fresh response for: https://pypi.org/simple/fromager/
        x No solution found when resolving dependencies:
        ╰─▶ Because you require fromager==1.0 and fromager==2.0, we can conclude that your requirements are unsatisfiable.
    """)
    match = build_environment._uv_missing_dependency_pattern.search(msg)
    assert match is not None



from packaging.requirements import Requirement
from packaging.version import Version

from fromager.commands.find_updates import _get_constraint_version


class TestGetConstraintVersion:
    """Test cases for _get_constraint_version function."""

    def test_equality_constraint(self):
        """Test that equality constraints return the correct version."""
        req = Requirement("package==1.2.3")
        result = _get_constraint_version(req)
        assert result == Version("1.2.3")

    def test_less_than_constraint(self):
        """Test that less than constraints return the correct version."""
        req = Requirement("package<2.0.0")
        result = _get_constraint_version(req)
        assert result == Version("2.0.0")

    def test_less_than_or_equal_constraint(self):
        """Test that less than or equal constraints return the correct version."""
        req = Requirement("package<=1.5.0")
        result = _get_constraint_version(req)
        assert result == Version("1.5.0")

    def test_greater_than_constraint_unsupported(self):
        """Test that greater than constraints return None (unsupported)."""
        req = Requirement("package>1.0.0")
        result = _get_constraint_version(req)
        assert result is None

    def test_greater_than_or_equal_constraint_unsupported(self):
        """Test that greater than or equal constraints return None (unsupported)."""
        req = Requirement("package>=1.0.0")
        result = _get_constraint_version(req)
        assert result is None

    def test_no_version_constraint(self):
        """Test that requirements without version constraints return None."""
        req = Requirement("package")
        result = _get_constraint_version(req)
        assert result is None

    def test_multiple_constraints_with_one_supported(self):
        """Test that multiple constraints with one supported constraint return the supported one."""
        req = Requirement("package>=1.0.0,<2.0.0")
        result = _get_constraint_version(req)
        assert result == Version("2.0.0")

    def test_constraint_with_markers(self):
        """Test that constraints with environment markers still work correctly."""
        req = Requirement("package==1.2.3; python_version >= '3.8'")
        result = _get_constraint_version(req)
        assert result == Version("1.2.3")

    def test_tilde_equal_constraint_supported(self):
        """Test that tilde equal constraints (~=) return the correct version."""
        req = Requirement("package~=1.4.2")
        result = _get_constraint_version(req)
        assert result == Version("1.4.2")

    def test_not_equal_constraint_supported(self):
        """Test that not equal constraints (!=) return the correct version."""
        req = Requirement("package!=1.0.0")
        result = _get_constraint_version(req)
        assert result == Version("1.0.0")

    def test_arbitrary_equality_constraint_supported(self):
        """Test that arbitrary equality constraints (===) return the correct version."""
        req = Requirement("package===1.0.0")
        result = _get_constraint_version(req)
        assert result == Version("1.0.0")

    def test_invalid_version_format(self):
        """Test that invalid version formats return None."""
        # We need to test the case where Version() construction fails
        # Let's create a requirement and then mock the Version parsing
        import unittest.mock

        req = Requirement("package==1.0.0")

        with unittest.mock.patch(
            "fromager.commands.find_updates.Version"
        ) as mock_version:
            mock_version.side_effect = ValueError("Invalid version")
            result = _get_constraint_version(req)
            assert result is None



import json
import pathlib
import re

from click.testing import CliRunner

from fromager.__main__ import main as fromager


def _extract_json_from_output(output: str) -> str:
    """Extract JSON content from output that may contain log messages."""
    # Find JSON array in the output
    json_match = re.search(r"\[\s*\{.*\}\s*\]", output, re.DOTALL)
    if json_match:
        return json_match.group(0)
    return "[]"


def _extract_csv_from_output(output: str) -> str:
    """Extract CSV content from output that may contain log messages."""
    lines = output.strip().split("\n")
    # Find lines that look like CSV (contain quotes and commas)
    csv_lines = [line for line in lines if '"' in line and "," in line]
    return "\n".join(csv_lines)


def test_list_overrides_basic(
    testdata_path: pathlib.Path, cli_runner: CliRunner
) -> None:
    """Test basic list-overrides functionality without --details."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"

    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
        ],
    )
    assert result.exit_code == 0
    assert "test-other-pkg" in result.stdout
    assert "test-pkg" in result.stdout
    assert "test-pkg-library" in result.stdout


def test_list_overrides_details_table(
    testdata_path: pathlib.Path, cli_runner: CliRunner
) -> None:
    """Test list-overrides --details with table format (default)."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"

    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--details",
        ],
    )
    assert result.exit_code == 0
    assert "Package Overrides" in result.stdout
    assert "test-other-pkg" in result.stdout
    assert "test-pkg" in result.stdout
    assert "test-pkg-library" in result.stdout


def test_list_overrides_details_json(
    testdata_path: pathlib.Path, cli_runner: CliRunner
) -> None:
    """Test list-overrides --details --format json."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"

    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--details",
            "--format",
            "json",
        ],
    )
    assert result.exit_code == 0

    # Extract JSON from output (filtering out log messages)
    json_output = _extract_json_from_output(result.stdout)
    data = json.loads(json_output)
    assert isinstance(data, list)
    assert len(data) == 3

    # Check that we have the expected packages
    packages = [item["package"] for item in data]
    assert "test-other-pkg" in packages
    assert "test-pkg" in packages
    assert "test-pkg-library" in packages

    # Check structure of first item
    first_item = data[0]
    assert "package" in first_item
    assert "version" in first_item
    assert "patches" in first_item
    assert "plugin_hooks" in first_item
    assert "rocm" in first_item  # variant column


def test_list_overrides_details_csv(
    testdata_path: pathlib.Path, cli_runner: CliRunner
) -> None:
    """Test list-overrides --details --format csv."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"

    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--details",
            "--format",
            "csv",
        ],
    )
    assert result.exit_code == 0

    # Extract CSV from output (filtering out log messages)
    csv_output = _extract_csv_from_output(result.stdout)
    lines = csv_output.strip().split("\n")
    assert len(lines) == 4  # header + 3 data rows

    # Check header
    header = lines[0]
    assert '"package"' in header
    assert '"version"' in header
    assert '"patches"' in header
    assert '"plugin_hooks"' in header
    assert '"rocm"' in header  # variant column

    # Check data rows
    assert any("test-other-pkg" in line for line in lines[1:])
    assert any("test-pkg" in line for line in lines[1:])
    assert any("test-pkg-library" in line for line in lines[1:])


def test_list_overrides_output_file(
    testdata_path: pathlib.Path, cli_runner: CliRunner, tmp_path: pathlib.Path
) -> None:
    """Test list-overrides with output file."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"
    output_file = tmp_path / "output.json"

    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--details",
            "--format",
            "json",
            "--output",
            str(output_file),
        ],
    )
    assert result.exit_code == 0
    # stdout may contain log messages, but should not contain the JSON data
    assert "test-other-pkg" not in result.stdout
    assert "test-pkg" not in result.stdout
    assert "test-pkg-library" not in result.stdout

    # Check that file was created and contains valid JSON
    assert output_file.exists()
    data = json.loads(output_file.read_text())
    assert isinstance(data, list)
    assert len(data) == 3


def test_list_overrides_format_option_help(cli_runner: CliRunner) -> None:
    """Test that the format option shows the correct choices in help."""
    result = cli_runner.invoke(
        fromager,
        ["list-overrides", "--help"],
    )
    assert result.exit_code == 0
    assert "--format [table|csv|json]" in result.stdout
    assert "requires --details" in result.stdout


def test_list_overrides_warnings_without_details(
    testdata_path: pathlib.Path, cli_runner: CliRunner
) -> None:
    """Test that warnings are shown when using format/output without details."""
    settings_file = testdata_path / "context" / "overrides" / "settings.yaml"
    patches_dir = testdata_path / "context" / "overrides" / "patches"

    # Test format warning
    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--format",
            "json",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (
        "Warning: --format option is ignored when --details is not used"
        in result.output
    )
    assert "test-other-pkg" in result.output

    # Test output warning
    result = cli_runner.invoke(
        fromager,
        [
            "--settings-file",
            str(settings_file),
            "--patches-dir",
            str(patches_dir),
            "list-overrides",
            "--output",
            "test.txt",
        ],
        catch_exceptions=False,
    )
    assert result.exit_code == 0
    assert (
        "Warning: --output option is ignored when --details is not used"
        in result.output
    )
    assert "test-other-pkg" in result.output



import logging
import pathlib
import sys

import click
from packaging.requirements import InvalidRequirement, Requirement

from fromager import bootstrapper, context, progress, requirements_file
from fromager.log import requirement_ctxvar
from fromager.requirements_file import RequirementType

logger = logging.getLogger(__name__)


@click.command()
@click.argument(
    "input_files_path",
    nargs=-1,
    required=True,
    type=click.Path(exists=False, path_type=pathlib.Path),
)
@click.option(
    "--resolve-requirements/--no-resolve-requirements",
    default=False,
    help="Resolve requirement and fail if a package or version cannot be resolved",
    show_default=True,
)
@click.pass_obj
def lint_requirements(
    wkctx: context.WorkContext,
    resolve_requirements: bool,
    input_files_path: list[pathlib.Path],
) -> None:
    """
    Command to lint the constraints.txt and requirements.txt files
    This command takes a single wildcard path string for constraints.txt and requirements.txt.
    It checks the formatting of these files and reports issues if found. Files with names that
    end with constraints.txt (e.g. constraints.txt, global-constraints.txt, etc.) are not allowed
    to contain extra dependencies. Additionally, it resolves valid input requirements to ensure
    we can find a matching version of each package.
    """

    if len(input_files_path) == 0:
        logger.error("no constraints.txt or requirements.txt found in given paths")
        sys.exit(1)

    failures: list[str] = []

    # Create bootstrapper for requirement resolution
    bt = bootstrapper.Bootstrapper(
        ctx=wkctx,
        progressbar=progress.Progressbar(None),
        prev_graph=None,
        cache_wheel_server_url=None,
        sdist_only=True,
    )

    for path in input_files_path:
        is_constraints: bool = path.name.endswith("constraints.txt")
        parsed_lines = requirements_file.parse_requirements_file(path)
        unique_entries: dict[str, Requirement] = {}
        for line in parsed_lines:
            try:
                requirement = Requirement(line)
                if requirement.name in unique_entries:
                    raise InvalidRequirement(
                        f"Duplicate entry, first found: {unique_entries[requirement.name]}"
                    )
                unique_entries[requirement.name] = requirement
                if is_constraints:
                    if requirement.extras:
                        raise InvalidRequirement(
                            f"{requirement.name}: Constraints files cannot contain extra dependencies"
                        )
                    if not requirement.specifier:
                        raise InvalidRequirement(
                            f"{requirement.name}: Constraints must have a version specifier"
                        )
            except InvalidRequirement as err:
                msg = f"{path}: {line}: {err}"
                logger.error(msg)
                failures.append(msg)

            # Resolve the requirement to ensure it can be found
            # Skip resolution for constraints files as they should only specify versions
            if resolve_requirements and not is_constraints:
                token = requirement_ctxvar.set(requirement)
                try:
                    _, version = bt.resolve_version(
                        req=requirement,
                        req_type=RequirementType.TOP_LEVEL,
                    )
                    logger.info(f"{requirement} resolves to {version}")
                except Exception as err:
                    logger.error(
                        f"{path}: {line}: Failed to resolve requirement: {err}"
                    )
                    failures.append(f"{path}: {line}: {err}")
                finally:
                    requirement_ctxvar.reset(token)

    if failures:
        click.echo("Validation error:", err=True)
        for failure in failures:
            click.echo(f" - {failure}", err=True)
        click.echo(
            f"ERROR: {len(failures)} failure(s) while validating {len(input_files_path)} file(s).",
            err=True,
        )
        sys.exit(1)
    else:
        click.echo(f"Resolve requirements: {resolve_requirements}")
        click.echo(f"Successfully validated {len(input_files_path)} file(s).")



import io
import pathlib
import textwrap
from unittest.mock import Mock, patch

import pytest
from click.testing import CliRunner
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import bootstrapper, context, dependency_graph, packagesettings
from fromager.commands import bootstrap


def test_get_requirements_single_arg():
    requirements = bootstrap._get_requirements_from_args(["a"], [])
    assert [Requirement("a")] == requirements


def test_get_requirements_multiple_args():
    requirements = bootstrap._get_requirements_from_args(["a", "b"], [])
    assert [Requirement("a"), Requirement("b")] == requirements


def test_get_requirements_args_and_file(tmp_path: pathlib.Path):
    requirements_file = tmp_path / "requirements.txt"
    requirements_file.write_text("c\n")
    requirements = bootstrap._get_requirements_from_args(
        ["a", "b"], [str(requirements_file)]
    )
    assert [
        Requirement("a"),
        Requirement("b"),
        Requirement("c"),
    ] == requirements


def test_ignore_based_on_marker():
    requirements = bootstrap._get_requirements_from_args(
        ['foo; python_version<"3.9"'], []
    )
    assert [] == requirements


def test_write_constraints_file_simple():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c<4.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        b==2.0
        c==3.0
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_resolvable_duplicate():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c<4.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [{"key": "c==3.1", "req_type": "install", "req": "c>3.0"}],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
        "c==3.1": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        b==2.0
        # NOTE: fromager selected c==3.1 from: ['3.0', '3.1']
        c==3.1
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_unresolvable_duplicate():
    """Test that unresolvable duplicates cause the function to return False and not write conflicting constraints.

    This test has conflicting requirements for package 'c':
    - a==1.0 requires c==3.0 (exact version 3.0)
    - b==2.0 requires c>3.0 (greater than 3.0)
    These cannot be satisfied simultaneously, so the function should return False
    and only write the packages that CAN be resolved.
    """
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c==3.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [{"key": "c==3.1", "req_type": "install", "req": "c>3.0"}],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
        "c==3.1": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)

    # Should return False due to impossible constraints for package 'c'
    result = bootstrap.write_constraints_file(graph, buffer)
    assert result is False

    # Should only write packages that CAN be resolved (no conflicting 'c' versions)
    output = buffer.getvalue()
    assert "a==1.0" in output
    assert "b==2.0" in output
    assert "c==" not in output  # No conflicting c versions should be written


def test_write_constraints_file_duplicates():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [
                {"key": "a==1.0", "req_type": "install", "req": "a"},
                {"key": "d==1.0", "req_type": "install", "req": "d"},
            ],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "c==3.0", "req_type": "install", "req": "c<=3.0"},
            ],
        },
        "d==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "c==3.1", "req_type": "install", "req": "c>=3.0"},
            ],
        },
        "c==3.0": {  # transformers 4.46
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [{"key": "b==2.0", "req_type": "install", "req": "b<2.1,>=2.0"}],
        },
        "c==3.1": {  # transformers 4.47
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [{"key": "b==2.1", "req_type": "install", "req": "b<2.2,>=2.1"}],
        },
        "b==2.0": {  # tokenizer
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [],
        },
        "b==2.1": {  # tokenizer
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.1",
            "canonicalized_name": "b",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)

    # This should return False because there's a genuine constraint conflict:
    # c==3.0 requires b<2.1,>=2.0 (only b==2.0 satisfies this)
    # c==3.1 requires b<2.2,>=2.1 (only b==2.1 satisfies this)
    # No single version of b can satisfy both constraints
    result = bootstrap.write_constraints_file(graph, buffer)
    assert result is False

    # When there are conflicts, no constraints should be written to the output
    output_content = buffer.getvalue()
    # Should contain resolved packages that don't have conflicts
    assert "a==1.0" in output_content
    assert "d==1.0" in output_content
    # c gets resolved to 3.0 before the conflict with b is detected
    assert "c==3.0" in output_content
    # Should NOT contain conflicted packages
    assert "b==" not in output_content


def test_write_constraints_file_multiples():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [
                {
                    "key": "a==2.7.0",
                    "req_type": "toplevel",
                    "req": "a==2.7.0",
                },
                {
                    "key": "b==0.26.1",
                    "req_type": "toplevel",
                    "req": "b==0.26.1",
                },
            ],
        },
        "b==0.26.1": {
            "download_url": "",
            "pre_built": False,
            "version": "0.26.1",
            "canonicalized_name": "b",
            "edges": [],
        },
        "b==0.26.2": {
            "download_url": "",
            "pre_built": False,
            "version": "0.26.2",
            "canonicalized_name": "b",
            "edges": [],
        },
        "a==2.7.0": {
            "download_url": "",
            "pre_built": False,
            "version": "2.7.0",
            "canonicalized_name": "a",
            "edges": [
                {
                    "key": "b==0.26.2",
                    "req_type": "install",
                    "req": "b<0.27.0,>=0.26.1",
                },
            ],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    assert bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==2.7.0
        # NOTE: fromager selected b==0.26.2 from: ['0.26.1', '0.26.2']
        b==0.26.2
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_prevents_false_resolution():
    """Test that packages marked as unresolvable in early iterations stay unresolvable.

    This test validates the fix for the bug where a package could appear unresolvable
    in iteration 1 (due to constraints from multiple versions), but then appear
    resolvable in iteration 2+ when user counts change after other packages resolve.

    Without the fix, 'conflicted' would be incorrectly resolved in a later iteration.
    With the fix, it should remain unresolvable throughout all iterations.
    """
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [
                {"key": "easy==1.0", "req_type": "toplevel", "req": "easy==1.0"},
                {"key": "pkg-a==1.0", "req_type": "toplevel", "req": "pkg-a==1.0"},
                {"key": "pkg-b==1.0", "req_type": "toplevel", "req": "pkg-b==1.0"},
            ],
        },
        "easy==1.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "easy",
            "edges": [],  # No dependencies, resolves immediately
        },
        "pkg-a==1.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "pkg-a",
            "edges": [
                {
                    "key": "conflicted==1.0",
                    "req_type": "install",
                    "req": "conflicted<1.5",
                },
                {
                    "key": "intermediate==1.0",
                    "req_type": "install",
                    "req": "intermediate",
                },
            ],
        },
        "pkg-b==1.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "pkg-b",
            "edges": [
                {
                    "key": "conflicted==2.0",
                    "req_type": "install",
                    "req": "conflicted>=2.0",
                },
            ],
        },
        "intermediate==1.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "intermediate",
            "edges": [
                {
                    "key": "conflicted==1.0",
                    "req_type": "install",
                    "req": "conflicted<1.5",
                },
            ],
        },
        "conflicted==1.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "conflicted",
            "edges": [],
        },
        "conflicted==2.0": {
            "download_url": "url",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "conflicted",
            "edges": [],
        },
    }

    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    result = bootstrap.write_constraints_file(graph, buffer)

    # Should return False because 'conflicted' has impossible constraints:
    # - pkg-a and intermediate need conflicted<1.5 (only 1.0 satisfies)
    # - pkg-b needs conflicted>=2.0 (only 2.0 satisfies)
    # - No version satisfies both <1.5 AND >=2.0
    assert result is False

    # Verify that packages that CAN be resolved are written
    output_content = buffer.getvalue()
    assert "easy==1.0" in output_content
    assert "pkg-a==1.0" in output_content
    assert "pkg-b==1.0" in output_content
    assert "intermediate==1.0" in output_content

    # Verify that the conflicted package is NOT written
    assert "conflicted==" not in output_content


def test_to_constraints_command_no_file_on_failure(tmp_path):
    """Test that the to-constraints logic doesn't create output files when there are constraint conflicts.

    This is a regression test for the bug where constraint resolution would write partial
    output even when it ultimately failed.
    """
    import io

    from fromager import dependency_graph

    # Create a graph with conflicts (same as test_write_constraints_file_unresolvable_duplicate)
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c==3.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [{"key": "c==3.1", "req_type": "install", "req": "c>3.0"}],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
        "c==3.1": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [],
        },
    }

    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)

    # Test the new to-constraints logic: use buffer first, only create file on success
    output_file = tmp_path / "constraints.txt"

    # Simulate the fixed to-constraints behavior
    buffer = io.StringIO()
    result = bootstrap.write_constraints_file(graph, buffer)

    # Should return False due to conflicts
    assert result is False

    # Because result is False, the output file should NOT be created
    # (In the old buggy behavior, the file would have been created with partial content)

    # Since result is False, we shouldn't write to the actual file
    # This simulates the fix where we check the result before creating the output file
    if result:
        with open(output_file, "w") as f:
            f.write(buffer.getvalue())

    # Verify the output file was NOT created
    assert not output_file.exists(), (
        f"Output file {output_file} should not have been created when constraint resolution failed"
    )


def test_skip_constraints_cli_option():
    """Test that the --skip-constraints option is available in the CLI"""
    runner = CliRunner()
    result = runner.invoke(bootstrap.bootstrap, ["--help"])

    # Check that the help text includes our new option
    assert "--skip-constraints" in result.output
    assert "Skip generating constraints.txt file" in result.output


@patch("fromager.gitutils.git_clone")
def test_resolve_version_from_git_url_with_submodules_enabled(
    mock_git_clone: Mock,
    tmp_context: context.WorkContext,
):
    """Test that git_clone is called with submodules=True when configured."""
    req = Requirement("test-pkg @ git+https://github.com/example/repo.git")

    mock_git_options = packagesettings.GitOptions(submodules=True)

    with patch.object(tmp_context, "package_build_info") as mock_pbi:
        mock_pbi_instance = Mock()
        mock_pbi_instance.git_options = mock_git_options
        mock_pbi.return_value = mock_pbi_instance

        with patch(
            "fromager.bootstrapper.Bootstrapper._get_version_from_package_metadata"
        ) as mock_get_version:
            with patch("shutil.move"):
                with patch("pathlib.Path.mkdir"):
                    mock_get_version.return_value = Version("1.0.0")

                    # Execute
                    bs = bootstrapper.Bootstrapper(tmp_context)
                    try:
                        bs._resolve_version_from_git_url(req)
                    except AssertionError:
                        # Expected since we're mocking everything
                        pass

    mock_git_clone.assert_called_once()
    call_args = mock_git_clone.call_args
    assert call_args.kwargs["submodules"] is True
    assert call_args.kwargs["repo_url"] == "https://github.com/example/repo.git"
    assert call_args.kwargs["ref"] is None


@patch("fromager.gitutils.git_clone")
def test_resolve_version_from_git_url_with_specific_submodule_paths(
    mock_git_clone: Mock,
    tmp_context: context.WorkContext,
):
    """Test that git_clone is called with specific submodule paths when configured."""
    req = Requirement("test-pkg @ git+https://github.com/example/repo.git")

    mock_git_options = packagesettings.GitOptions(
        submodule_paths=["vendor/lib1", "vendor/lib2"]
    )

    with patch.object(tmp_context, "package_build_info") as mock_pbi:
        mock_pbi_instance = Mock()
        mock_pbi_instance.git_options = mock_git_options
        mock_pbi.return_value = mock_pbi_instance

        with patch(
            "fromager.bootstrapper.Bootstrapper._get_version_from_package_metadata"
        ) as mock_get_version:
            with patch("shutil.move"):
                with patch("pathlib.Path.mkdir"):
                    mock_get_version.return_value = Version("1.0.0")

                    bs = bootstrapper.Bootstrapper(tmp_context)
                    try:
                        bs._resolve_version_from_git_url(req)
                    except AssertionError:
                        # Expected since we're mocking everything
                        pass

    mock_git_clone.assert_called_once()
    call_args = mock_git_clone.call_args
    assert call_args.kwargs["submodules"] == ["vendor/lib1", "vendor/lib2"]


@patch("fromager.gitutils.git_clone")
def test_resolve_version_from_git_url_with_submodules_disabled(
    mock_git_clone: Mock,
    tmp_context: context.WorkContext,
):
    """Test that git_clone is called with submodules=False by default."""
    req = Requirement("test-pkg @ git+https://github.com/example/repo.git")

    with patch(
        "fromager.bootstrapper.Bootstrapper._get_version_from_package_metadata"
    ) as mock_get_version:
        with patch("shutil.move"):
            with patch("pathlib.Path.mkdir"):
                mock_get_version.return_value = Version("1.0.0")

                bs = bootstrapper.Bootstrapper(tmp_context)
                try:
                    bs._resolve_version_from_git_url(req)
                except AssertionError:
                    # Expected since we're mocking everything
                    pass

    mock_git_clone.assert_called_once()
    call_args = mock_git_clone.call_args
    assert call_args.kwargs["submodules"] is False


@patch("fromager.gitutils.git_clone")
def test_resolve_version_from_git_url_with_git_ref(
    mock_git_clone: Mock,
    tmp_context: context.WorkContext,
):
    """Test that git_clone is called with the correct ref when URL includes @ref."""
    req = Requirement("test-pkg @ git+https://github.com/example/repo.git@v1.2.3")

    mock_git_options = packagesettings.GitOptions(submodules=True)

    with patch.object(tmp_context, "package_build_info") as mock_pbi:
        mock_pbi_instance = Mock()
        mock_pbi_instance.git_options = mock_git_options
        mock_pbi.return_value = mock_pbi_instance

        with patch(
            "fromager.bootstrapper.Bootstrapper._get_version_from_package_metadata"
        ) as mock_get_version:
            with patch("shutil.move"):
                with patch("pathlib.Path.mkdir"):
                    mock_get_version.return_value = Version("1.2.3")

                    bs = bootstrapper.Bootstrapper(tmp_context)
                    try:
                        bs._resolve_version_from_git_url(req)
                    except AssertionError:
                        # Expected since we're mocking everything
                        pass

    mock_git_clone.assert_called_once()
    call_args = mock_git_clone.call_args
    assert call_args.kwargs["submodules"] is True
    assert call_args.kwargs["repo_url"] == "https://github.com/example/repo.git"
    assert call_args.kwargs["ref"] == "v1.2.3"


def test_resolve_version_from_git_url_invalid_scheme(tmp_context: context.WorkContext):
    """Test that non-git URLs raise ValueError."""
    req = Requirement("test-pkg @ https://github.com/example/repo.git")

    bs = bootstrapper.Bootstrapper(tmp_context)
    with pytest.raises(ValueError, match="unable to handle URL scheme"):
        bs._resolve_version_from_git_url(req)



import logging
from enum import Enum

import click
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import context, overrides, resolver

logger = logging.getLogger(__name__)


class DistributionType(Enum):
    """Distribution type for version lookup"""

    DEFAULT = "default"
    SDIST = "sdist"
    WHEEL = "wheel"
    BOTH = "both"


@click.command()
@click.option(
    "--distribution-type",
    type=click.Choice(DistributionType, case_sensitive=False),
    default=DistributionType.DEFAULT.value,
    help="Distribution type to include in version lookup (default: use package settings, sdist: source only, wheel: wheels only, both: include both sdists and wheels)",
)
@click.option(
    "--sdist-server-url",
    default=resolver.PYPI_SERVER_URL,
    help="URL to the Python package index to use for version lookup",
)
@click.option(
    "--ignore-no-versions/--no-ignore-no-versions",
    default=False,
    help="Do not treat missing versions as an error",
)
@click.option(
    "--format-as-requirements/--no-format-as-requirements",
    default=False,
    help="Format output as requirement specifiers (name==version) instead of just version numbers",
)
@click.argument("requirement_spec", required=True)
@click.pass_obj
def list_versions(
    wkctx: context.WorkContext,
    requirement_spec: str,
    distribution_type: str,
    sdist_server_url: str,
    ignore_no_versions: bool,
    format_as_requirements: bool,
) -> None:
    """List all available versions for a package requirement specifier.

    The REQUIREMENT_SPEC should be a package requirement specification like:
    - "package_name" (any version)
    - "package_name>=1.0" (versions >= 1.0)
    - "package_name==1.*" (versions matching 1.*)

    This command uses the get_resolver_provider hook to retrieve a resolver
    provider for the package in case there is a custom provider configured.

    Distribution types:
    - "default": Use package settings for include_sdists/include_wheels
    - "sdist": Only include source distributions
    - "wheel": Only include wheels
    - "both": Include both source distributions and wheels
    """
    try:
        req = Requirement(requirement_spec)
    except Exception as e:
        raise click.ClickException(
            f"Invalid requirement specification '{requirement_spec}': {e}"
        ) from e

    pbi = wkctx.package_build_info(req)
    override_sdist_server_url = pbi.resolver_sdist_server_url(sdist_server_url)

    # Determine include flags based on distribution type
    dist_type = DistributionType(distribution_type)
    match dist_type:
        case DistributionType.SDIST:
            include_sdists = True
            include_wheels = False
        case DistributionType.WHEEL:
            include_sdists = False
            include_wheels = True
        case DistributionType.BOTH:
            include_sdists = True
            include_wheels = True
        case _:  # DEFAULT
            # Use package settings defaults
            package_settings = wkctx.settings.package_setting(req.name)
            include_sdists = package_settings.resolver_dist.include_sdists
            include_wheels = package_settings.resolver_dist.include_wheels

    logger.info(f"Looking up versions for {req.name}")
    if req.specifier:
        logger.info(f"Filtering versions with specifier: {req.specifier}")
    logger.info(
        f"Using distribution type: {dist_type.value} (sdists: {include_sdists}, wheels: {include_wheels})"
    )

    provider = overrides.find_and_invoke(
        req.name,
        "get_resolver_provider",
        resolver.default_resolver_provider,
        ctx=wkctx,
        req=req,
        include_sdists=include_sdists,
        include_wheels=include_wheels,
        sdist_server_url=override_sdist_server_url,
    )

    # Get all available candidates from the provider
    candidates = list(
        provider.find_matches(
            identifier=req.name,
            requirements={req.name: [req]},
            incompatibilities={req.name: []},
        )
    )

    if not candidates:
        if ignore_no_versions:
            logger.warning(f"No versions found for {req.name}")
            return
        else:
            raise click.ClickException(f"No versions found for {req.name}")

    versions: list[Version] = sorted(set(candidate.version for candidate in candidates))
    logger.info(f"Found {len(versions)} version(s)")

    for version in versions:
        if format_as_requirements:
            print(f"{req.name}=={version}")
        else:
            print(version)



import logging
import pathlib

from packaging.requirements import Requirement
from packaging.version import Version

from fromager import context, sources, build_environment

logger = logging.getLogger(__name__)


def update_extra_environ(
    *,
    ctx: context.WorkContext,
    req: Requirement,
    version: Version | None,
    sdist_root_dir: pathlib.Path,
    extra_environ: dict[str, str],
    build_env: build_environment.BuildEnvironment,
) -> None:
    """Update extra_environ in-place"""
    logger.info("update_extra_environ resolved_version=%s", version)
    marker = ctx.work_dir / "update_extra_environ.txt"
    with marker.open(encoding="utf-8", mode="a") as f:
        f.write(f"{version}\n")
    return None


def build_sdist(
    ctx: context.WorkContext,
    extra_environ: dict,
    req: Requirement,
    sdist_root_dir: pathlib.Path,
    version: Version,
    build_env: build_environment.BuildEnvironment,
) -> pathlib.Path:
    return sources.pep517_build_sdist(
        ctx=ctx,
        extra_environ=extra_environ,
        req=req,
        sdist_root_dir=sdist_root_dir,
        version=version,
        build_env=build_env,
    )



import contextlib
import contextvars
import logging
import threading
import typing

from packaging.requirements import Requirement
from packaging.version import Version

TERSE_LOG_FMT = "%(asctime)s %(levelname)s %(message)s"
TERSE_DATE_FMT = "%H:%M:%S"
VERBOSE_LOG_FMT = "%(asctime)s %(levelname)s:%(name)s:%(lineno)d: %(message)s"

requirement_ctxvar: contextvars.ContextVar[Requirement] = contextvars.ContextVar(
    "requirement"
)
version_ctxvar: contextvars.ContextVar[Version] = contextvars.ContextVar("version")


@contextlib.contextmanager
def req_ctxvar_context(
    req: Requirement, version: Version | None = None
) -> typing.Generator[None, None, None]:
    """Context manager for requirement_ctxvar and version_ctxvar"""
    req_token = requirement_ctxvar.set(req)
    if version is not None:
        version_token = version_ctxvar.set(version)
    try:
        yield None
    finally:
        if version is not None:
            version_ctxvar.reset(version_token)
        requirement_ctxvar.reset(req_token)


class FromagerLogRecord(logging.LogRecord):
    """Logger record factory to add requirement and version from context var

    The class prepends f"{req.name}: " to every log message if-and-only-if
    ``requirement_ctxvar`` is set for the current context. The context var
    must be set at the beginning of a new requirement scope and reset at the
    end of a scope.

    If the ``version_ctxvar`` is also set, then the log record is prepended with
    f"{req.name}-{version}".

    ::
        for req in reqs:
            token = requirement_ctxvar.set(req)
            do_stuff(req)
            requirement_ctxvar.reset(token)
    """

    def getMessage(self) -> str:  # noqa: N802
        msg = super().getMessage()
        try:
            req = requirement_ctxvar.get()
        except LookupError:
            return msg
        else:
            try:
                version = version_ctxvar.get()
            except LookupError:
                return f"{req.name}: {msg}"
            else:
                return f"{req.name}-{version}: {msg}"


class ThreadLogFilter(logging.Filter):
    """Filter that only emits records for the given thread name"""

    def __init__(self, thread_name: str):
        self._thread_name = thread_name

    def filter(self, record: logging.LogRecord) -> bool:
        return threading.current_thread().name == self._thread_name
